#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock

from neutron.common.ovn import constants as ovn_const
from neutron.objects import port_forwarding as port_forwarding_obj
from neutron.services.portforwarding.constants import PORT_FORWARDING
from neutron.services.portforwarding.constants import PORT_FORWARDING_PLUGIN
from neutron.services.portforwarding.drivers.ovn import driver \
    as port_forwarding
from neutron.tests import base
from neutron.tests.unit import fake_resources
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as const
from neutron_lib.plugins import constants as plugin_constants
from oslo_utils import uuidutils
from ovsdbapp import constants as ovsdbapp_const


class TestOVNPortForwardingBase(base.BaseTestCase):
    def setUp(self):
        super(TestOVNPortForwardingBase, self).setUp()
        self.context = mock.Mock()
        self.l3_plugin = mock.Mock()
        self.l3_plugin._nb_ovn = fake_resources.FakeOvsdbNbOvnIdl()
        self.txn = self.l3_plugin._nb_ovn.transaction

    def _fake_pf_obj(self, is_object=False, **kwargs):
        pf_obj_defaults_dict = {
            'id': 'id',
            'floatingip_id': 'fip_id',
            'protocol': 'udp',
            'floating_ip_address': 'fip_addr',
            'external_port': 'ext_port',
            'internal_ip_address': 'internal_addr',
            'internal_port': 'internal_port',
            'router_id': 'rtr_id'
        }
        pf_obj_dict = {**pf_obj_defaults_dict, **kwargs}
        if is_object:
            return port_forwarding_obj.PortForwarding(**pf_obj_dict)
        return mock.Mock(**pf_obj_dict)

    def _fake_pf_payload_entry(self, curr_pf_id, orig_pf_id=None, **kwargs):
        mock_pf_payload = mock.Mock()
        states = []
        if 'context' not in kwargs:
            mock_pf_payload.context = self.context
        if orig_pf_id:
            fake_pf_obj = self._fake_pf_obj(**kwargs)
            fake_pf_obj.floatingip_id = orig_pf_id
            states.append(fake_pf_obj)
        if curr_pf_id:
            fake_pf_obj = self._fake_pf_obj(**kwargs)
            mock_pf_payload.latest_state = fake_pf_obj
            mock_pf_payload.latest_state.floatingip_id = curr_pf_id
            states.append(fake_pf_obj)
        else:
            mock_pf_payload.latest_state = None

        mock_pf_payload.states = states
        return mock_pf_payload


class TestOVNPortForwardingHandler(TestOVNPortForwardingBase):
    def setUp(self):
        super(TestOVNPortForwardingHandler, self).setUp()
        self.handler = port_forwarding.OVNPortForwardingHandler()

    def test_get_lb_protocol(self):
        fake_pf_obj = self._fake_pf_obj(protocol='udp')
        self.assertEqual(ovsdbapp_const.PROTO_UDP,
                         self.handler._get_lb_protocol(fake_pf_obj))
        fake_pf_obj = self._fake_pf_obj(protocol='tcp')
        self.assertEqual(ovsdbapp_const.PROTO_TCP,
                         self.handler._get_lb_protocol(fake_pf_obj))
        fake_pf_obj = self._fake_pf_obj(protocol='xxx')
        self.assertRaises(KeyError, self.handler._get_lb_protocol,
                          fake_pf_obj)

    def test_lb_names(self):
        expected_names = ['pf-floatingip-id-udp', 'pf-floatingip-id-tcp']
        names = self.handler.lb_names('id')
        self.assertCountEqual(expected_names, names)

    def test_get_lb_attributes(self):
        fake_pf_obj = self._fake_pf_obj()
        lb_name, vip, internal_ip, rtr_name = self.handler._get_lb_attributes(
            fake_pf_obj)
        self.assertEqual('pf-floatingip-fip_id-udp', lb_name)
        self.assertEqual('fip_addr:ext_port', vip)
        self.assertCountEqual(['internal_addr:internal_port'], internal_ip)
        self.assertEqual('neutron-rtr_id', rtr_name)

    @mock.patch.object(port_forwarding.LOG, 'info')
    def test_port_forwarding_created(self, m_info):
        fake_pf_obj = self._fake_pf_obj(
            is_object=True,
            id=uuidutils.generate_uuid(),
            floatingip_id=uuidutils.generate_uuid(),
            router_id=uuidutils.generate_uuid(),
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=2222,
            internal_ip_address='192.168.0.1',
            internal_port=22,
        )
        exp_lb_name, exp_vip, exp_internal_ips, exp_rtr_name = (
            self.handler._get_lb_attributes(fake_pf_obj))
        exp_protocol = self.handler._get_lb_protocol(fake_pf_obj)
        self.handler.port_forwarding_created(
            self.txn, self.l3_plugin._nb_ovn, fake_pf_obj)
        info_args, _info_kwargs = m_info.call_args_list[0]
        self.assertIn('CREATE for port-forwarding', info_args[0])
        self.assertEqual(2, len(self.txn.add.call_args_list))
        exp_external_ids = {
            ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY: PORT_FORWARDING_PLUGIN,
            ovn_const.OVN_FIP_EXT_ID_KEY: fake_pf_obj.floatingip_id,
            ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: exp_rtr_name,
        }
        self.l3_plugin._nb_ovn.lb_add.assert_called_once_with(
            exp_lb_name, exp_vip, exp_internal_ips, exp_protocol,
            may_exist=True, external_ids=exp_external_ids)
        self.l3_plugin._nb_ovn.lr_lb_add.assert_called_once_with(
            exp_rtr_name, exp_lb_name, may_exist=True)

    @mock.patch.object(port_forwarding.LOG, 'info')
    def test_port_forwarding_created_with_ranges(self, m_info):
        id = uuidutils.generate_uuid()
        floatingip_id = uuidutils.generate_uuid()
        router_id = uuidutils.generate_uuid()
        fake_pf_obj_ranges = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=None,
            external_port_range='2222:2223',
            internal_ip_address='192.168.0.1',
            internal_port=None,
            internal_port_range='22:23'
        )
        fake_pf_obj_1 = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=2222,
            external_port_range=None,
            internal_ip_address='192.168.0.1',
            internal_port=22,
            internal_port_range=None
        )
        fake_pf_obj_2 = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=2223,
            external_port_range=None,
            internal_ip_address='192.168.0.1',
            internal_port=23,
            internal_port_range=None
        )
        exp_lb_name_1, exp_vip_1, exp_internal_ips_1, exp_rtr_name_1 = (
            self.handler._get_lb_attributes(fake_pf_obj_1, True))
        exp_lb_name_2, exp_vip_2, exp_internal_ips_2, exp_rtr_name_2 = (
            self.handler._get_lb_attributes(fake_pf_obj_2, True))
        exp_protocol_1 = self.handler._get_lb_protocol(fake_pf_obj_1)
        exp_protocol_2 = self.handler._get_lb_protocol(fake_pf_obj_2)
        self.handler.port_forwarding_created(
            self.txn, self.l3_plugin._nb_ovn, fake_pf_obj_ranges)
        info_args, _info_kwargs = m_info.call_args_list[0]
        self.assertIn('CREATE for port-forwarding', info_args[0])
        self.assertEqual(4, len(self.txn.add.call_args_list))
        info_args, _info_kwargs = m_info.call_args_list[1]
        self.assertIn('CREATE for port-forwarding', info_args[0])
        self.assertEqual(4, len(self.txn.add.call_args_list))
        exp_external_ids_1 = {
            ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY: PORT_FORWARDING_PLUGIN,
            ovn_const.OVN_FIP_EXT_ID_KEY: fake_pf_obj_1.floatingip_id,
            ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: exp_rtr_name_1,
        }
        exp_external_ids_2 = {
            ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY: PORT_FORWARDING_PLUGIN,
            ovn_const.OVN_FIP_EXT_ID_KEY: fake_pf_obj_1.floatingip_id,
            ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: exp_rtr_name_2,
        }
        self.l3_plugin._nb_ovn.lb_add.assert_any_call(
            exp_lb_name_1, exp_vip_1, exp_internal_ips_1, exp_protocol_1,
            may_exist=True, external_ids=exp_external_ids_1)
        self.l3_plugin._nb_ovn.lr_lb_add.assert_any_call(
            exp_rtr_name_1, exp_lb_name_1, may_exist=True)
        self.l3_plugin._nb_ovn.lb_add.assert_any_call(
            exp_lb_name_2, exp_vip_2, exp_internal_ips_2, exp_protocol_2,
            may_exist=True, external_ids=exp_external_ids_2)
        self.l3_plugin._nb_ovn.lr_lb_add.assert_any_call(
            exp_rtr_name_2, exp_lb_name_2, may_exist=True)

    @mock.patch.object(port_forwarding.LOG, 'info')
    @mock.patch.object(
        port_forwarding.OVNPortForwardingHandler, '_port_forwarding_deleted')
    @mock.patch.object(
        port_forwarding.OVNPortForwardingHandler, '_port_forwarding_created')
    def test_port_forwarding_updated_with_ranges(self, m_created, m_deleted,
                                                 m_info):
        id = uuidutils.generate_uuid()
        floatingip_id = uuidutils.generate_uuid()
        router_id = uuidutils.generate_uuid()
        fake_pf_obj_ranges_old = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=None,
            external_port_range='2222:2223',
            internal_ip_address='192.168.0.1',
            internal_port=None,
            internal_port_range='22:23'
        )
        fake_pf_obj_1_old = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=2222,
            external_port_range=None,
            internal_ip_address='192.168.0.1',
            internal_port=22,
            internal_port_range=None
        )
        fake_pf_obj_2_old = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=2223,
            external_port_range=None,
            internal_ip_address='192.168.0.1',
            internal_port=23,
            internal_port_range=None
        )
        fake_pf_obj_ranges = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='tcp',
            floating_ip_address='10.0.0.1',
            external_port=None,
            external_port_range='2222:2223',
            internal_ip_address='192.168.0.1',
            internal_port=None,
            internal_port_range='22:23'
        )
        fake_pf_obj_1 = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='tcp',
            floating_ip_address='10.0.0.1',
            external_port=2222,
            external_port_range=None,
            internal_ip_address='192.168.0.1',
            internal_port=22,
            internal_port_range=None
        )
        fake_pf_obj_2 = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='tcp',
            floating_ip_address='10.0.0.1',
            external_port=2223,
            external_port_range=None,
            internal_ip_address='192.168.0.1',
            internal_port=23,
            internal_port_range=None
        )

        def do_simple_comparation(pf_1, pf_2):
            pf_1_dict = {f: getattr(pf_1, f)
                         for f in pf_1.fields if hasattr(pf_1, f)}
            pf_2_dict = {f: getattr(pf_2, f)
                         for f in pf_2.fields if hasattr(pf_2, f)}
            self.assertEqual(pf_1_dict, pf_2_dict)

        self.handler.port_forwarding_updated(
            self.txn, self.l3_plugin._nb_ovn, fake_pf_obj_ranges,
            fake_pf_obj_ranges_old)
        info_args, _info_kwargs = m_info.call_args_list[0]
        self.assertIn('UPDATE for port-forwarding', info_args[0])
        deleted_pf_1 = m_deleted.call_args_list[0][0][2]
        deleted_pf_2 = m_deleted.call_args_list[1][0][2]
        do_simple_comparation(deleted_pf_1, fake_pf_obj_1_old)
        do_simple_comparation(deleted_pf_2, fake_pf_obj_2_old)
        created_pf_1 = m_created.call_args_list[0][0][2]
        created_pf_2 = m_created.call_args_list[1][0][2]
        do_simple_comparation(created_pf_1, fake_pf_obj_1)
        do_simple_comparation(created_pf_2, fake_pf_obj_2)

    @mock.patch.object(port_forwarding.LOG, 'info')
    @mock.patch.object(
        port_forwarding.OVNPortForwardingHandler, '_port_forwarding_deleted')
    @mock.patch.object(
        port_forwarding.OVNPortForwardingHandler, '_port_forwarding_created')
    def test_port_forwarding_updated(self, m_created, m_deleted, m_info):
        fake_pf_obj = self._fake_pf_obj(
            is_object=True,
            id=uuidutils.generate_uuid(),
            floatingip_id=uuidutils.generate_uuid(),
            router_id=uuidutils.generate_uuid(),
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=2222,
            internal_ip_address='192.168.0.1',
            internal_port=22,
        )
        fake_orig_pf_obj = self._fake_pf_obj(
            is_object=True,
            id=uuidutils.generate_uuid(),
            floatingip_id=uuidutils.generate_uuid(),
            router_id=uuidutils.generate_uuid(),
            protocol='tcp',
            floating_ip_address='10.0.0.1',
            external_port=2222,
            internal_ip_address='192.168.0.1',
            internal_port=22,
        )
        self.handler.port_forwarding_updated(
            self.txn, self.l3_plugin._nb_ovn, fake_pf_obj, fake_orig_pf_obj)
        info_args, _info_kwargs = m_info.call_args_list[0]
        self.assertIn('UPDATE for port-forwarding', info_args[0])
        m_deleted.assert_called_once_with(self.txn, self.l3_plugin._nb_ovn,
                                          fake_orig_pf_obj, is_range=False)
        m_created.assert_called_once_with(self.txn, self.l3_plugin._nb_ovn,
                                          fake_pf_obj, is_range=False)

    @mock.patch.object(port_forwarding.LOG, 'info')
    def test_port_forwarding_deleted_with_ranges(self, m_info):
        id = uuidutils.generate_uuid()
        floatingip_id = uuidutils.generate_uuid()
        router_id = uuidutils.generate_uuid()
        fake_pf_obj_ranges = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=None,
            external_port_range='2222:2223',
            internal_ip_address='192.168.0.1',
            internal_port=None,
            internal_port_range='22:23'
        )
        fake_pf_obj_1 = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=2222,
            external_port_range=None,
            internal_ip_address='192.168.0.1',
            internal_port=22,
            internal_port_range=None
        )
        fake_pf_obj_2 = self._fake_pf_obj(
            is_object=True,
            id=id,
            floatingip_id=floatingip_id,
            router_id=router_id,
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=2223,
            external_port_range=None,
            internal_ip_address='192.168.0.1',
            internal_port=23,
            internal_port_range=None
        )

        exp_lb_name_1, exp_vip_1, _, _ = self.handler._get_lb_attributes(
            fake_pf_obj_1, True)
        exp_lb_name_2, exp_vip_2, _, _ = self.handler._get_lb_attributes(
            fake_pf_obj_2, True)
        self.handler.port_forwarding_deleted(
            self.txn, self.l3_plugin._nb_ovn, fake_pf_obj_ranges)
        info_args, _info_kwargs = m_info.call_args_list[0]
        self.assertIn('DELETE for port-forwarding', info_args[0])
        self.assertEqual(2, len(self.txn.add.call_args_list))
        self.l3_plugin._nb_ovn.lb_del.assert_any_call(
            exp_lb_name_1, exp_vip_1, if_exists=mock.ANY)
        self.l3_plugin._nb_ovn.lb_del.assert_any_call(
            exp_lb_name_2, exp_vip_2, if_exists=mock.ANY)

    @mock.patch.object(port_forwarding.LOG, 'info')
    def test_port_forwarding_deleted(self, m_info):
        fake_pf_obj = self._fake_pf_obj(
            is_object=True,
            id=uuidutils.generate_uuid(),
            floatingip_id=uuidutils.generate_uuid(),
            router_id=uuidutils.generate_uuid(),
            protocol='udp',
            floating_ip_address='10.0.0.1',
            external_port=2222,
            internal_ip_address='192.168.0.1',
            internal_port=22,
        )
        exp_lb_name, exp_vip, _, _ = self.handler._get_lb_attributes(
            fake_pf_obj)
        self.handler.port_forwarding_deleted(
            self.txn, self.l3_plugin._nb_ovn, fake_pf_obj)
        info_args, _info_kwargs = m_info.call_args_list[0]
        self.assertIn('DELETE for port-forwarding', info_args[0])
        self.assertEqual(1, len(self.txn.add.call_args_list))
        self.l3_plugin._nb_ovn.lb_del.assert_called_once_with(
            exp_lb_name, exp_vip, if_exists=mock.ANY)


class TestOVNPortForwarding(TestOVNPortForwardingBase):
    def setUp(self):
        super(TestOVNPortForwarding, self).setUp()
        self.pf_plugin = mock.Mock()
        self.handler = mock.Mock()
        get_mock_pf_plugin = lambda alias: self.pf_plugin if (
                alias == plugin_constants.PORTFORWARDING) else None
        self.fake_get_dir_object = mock.patch(
            "neutron_lib.plugins.directory.get_plugin",
            side_effect=get_mock_pf_plugin).start()
        self.fake_handler_object = mock.patch(
            "neutron.services.portforwarding.drivers.ovn.driver."
            "OVNPortForwardingHandler",
            return_value=self.handler).start()
        self._ovn_pf = port_forwarding.OVNPortForwarding(self.l3_plugin)
        self.mock_pf_revs = mock.Mock()
        self.fake_check_rev = mock.patch.object(
            self._ovn_pf, '_add_check_rev',
            return_value=self.mock_pf_revs).start()
        self.fake_db_rev = mock.patch.object(
            self._ovn_pf, '_do_db_rev_bump_revision').start()

    def test_init(self):
        self.assertIsNotNone(self._ovn_pf)
        self.assertEqual(self._ovn_pf._l3_plugin, self.l3_plugin)
        self.assertEqual(self._ovn_pf._handler, self.handler)
        self.assertEqual(self._ovn_pf._pf_plugin, self.pf_plugin)

    def test_register(self):
        with mock.patch.object(registry, 'subscribe') as mock_subscribe:
            self._ovn_pf.register(mock.ANY, mock.ANY, mock.Mock())
            calls = [mock.call.mock_subscribe(mock.ANY,
                                              PORT_FORWARDING,
                                              events.AFTER_CREATE),
                     mock.call.mock_subscribe(mock.ANY,
                                              resources.ROUTER_INTERFACE,
                                              events.AFTER_CREATE),
                     mock.call.mock_subscribe(mock.ANY,
                                              PORT_FORWARDING,
                                              events.AFTER_UPDATE),
                     mock.call.mock_subscribe(mock.ANY,
                                              PORT_FORWARDING,
                                              events.AFTER_DELETE),
                     mock.call.mock_subscribe(mock.ANY,
                                              resources.ROUTER_INTERFACE,
                                              events.AFTER_DELETE)]
            mock_subscribe.assert_has_calls(calls)

    def test_get_pf_objs(self):
        _uuid = uuidutils.generate_uuid
        fip_id = _uuid()
        fake_pf_dicts = [{'id': _uuid(),
                          'floatingip_id': fip_id,
                          'external_port': pf_port,
                          'protocol': 'tcp',
                          'internal_port_id': _uuid(),
                          'internal_ip_address': '1.1.1.2',
                          'internal_port': pf_port,
                          'floating_ip_address': '111.111.111.111',
                          'router_id': _uuid()} for pf_port in range(22, 32)]

        self.pf_plugin.get_floatingip_port_forwardings = mock.Mock(
            return_value=fake_pf_dicts)
        pf_objs = self._ovn_pf._get_pf_objs(self.context, fip_id)
        self.pf_plugin.get_floatingip_port_forwardings.assert_called_once_with(
            self.context, fip_id)
        for index, fake_pf_dict in enumerate(fake_pf_dicts):
            self.assertEqual(fake_pf_dict['id'], pf_objs[index].id)
            self.assertEqual(fake_pf_dict['floatingip_id'],
                             pf_objs[index].floatingip_id)
            self.assertEqual(fake_pf_dict['external_port'],
                             pf_objs[index].external_port)
            self.assertEqual(fake_pf_dict['internal_port_id'],
                             pf_objs[index].internal_port_id)
            self.assertEqual(fake_pf_dict['router_id'],
                             pf_objs[index].router_id)

    def test_get_fip_objs(self):
        payload = self._fake_pf_payload_entry(1, 2)
        self.l3_plugin.get_floatingip = lambda _, fip_id: fip_id * 10
        fip_objs = self._ovn_pf._get_fip_objs(self.context, payload)
        self.assertEqual({1: 10, 2: 20}, fip_objs)

    def _handle_notification_common(self, event_type, payload=None,
                                    fip_objs=None):
        if not payload:
            payload = []
        if not fip_objs:
            fip_objs = {}
        self.fake_db_rev.reset_mock()
        with mock.patch.object(self._ovn_pf, '_get_fip_objs',
                               return_value=fip_objs) as mock_get_fip_objs:
            self._ovn_pf._handle_notification(None, event_type,
                                              self.pf_plugin, payload)
            self.assertTrue(self.fake_db_rev.called or not fip_objs)
            if not payload:
                return
            mock_get_fip_objs.assert_called_once_with(self.context, payload)
            if fip_objs:
                calls = [
                    mock.call(mock.ANY, self.l3_plugin._nb_ovn, fip_id,
                              fip_obj)
                    for fip_id, fip_obj in fip_objs.items()]
                self.fake_check_rev.assert_has_calls(calls)
                self.fake_db_rev.assert_called_once_with(
                    self.context, self.mock_pf_revs)

    def test_handle_notification_noop(self):
        self._handle_notification_common(events.AFTER_CREATE)
        weird_event_type = 666
        fake_payload = self._fake_pf_payload_entry(None)
        self._handle_notification_common(weird_event_type, fake_payload)

    def test_handle_notification_basic(self):
        fake_payload_entry = self._fake_pf_payload_entry(1)
        self._handle_notification_common(events.AFTER_CREATE,
                                         fake_payload_entry)
        self.handler.port_forwarding_created.assert_called_once_with(
            mock.ANY, self.l3_plugin._nb_ovn, fake_payload_entry.latest_state)

    def test_handle_notification_create(self):
        fip_objs = {1: {'description': 'one'},
                    3: {'description': 'three', 'revision_number': '321'}}
        for id in range(1, 4):
            fake_payload = self._fake_pf_payload_entry(id)
            self._handle_notification_common(events.AFTER_CREATE,
                                             fake_payload,
                                             fip_objs)
            calls = [mock.call(mock.ANY, self.l3_plugin._nb_ovn,
                               fake_payload.latest_state)]
            self.handler.port_forwarding_created.assert_has_calls(calls)
            update_calls = [mock.call(
                self.context, fake_payload.latest_state.floatingip_id,
                const.FLOATINGIP_STATUS_ACTIVE)]
            self.l3_plugin.update_floatingip_status.assert_has_calls(
                update_calls)

    def test_handle_notification_update(self):
        fip_objs = {100: {'description': 'hundred'}, 101: {}}
        fake_payload = self._fake_pf_payload_entry(100, 100)
        self._handle_notification_common(events.AFTER_UPDATE, fake_payload,
                                         fip_objs)
        calls = [mock.call(mock.ANY, self.l3_plugin._nb_ovn,
                           fake_payload.latest_state,
                           fake_payload.states[0])]
        self.handler.port_forwarding_updated.assert_has_calls(calls)

        fake_payload = self._fake_pf_payload_entry(101, 101)
        self._handle_notification_common(events.AFTER_UPDATE, fake_payload,
                                         fip_objs)
        calls = [mock.call(mock.ANY, self.l3_plugin._nb_ovn,
                           fake_payload.latest_state,
                           fake_payload.states[0])]
        self.handler.port_forwarding_updated.assert_has_calls(calls)

    def test_handle_notification_delete(self):
        fip_objs = {1: {'description': 'one'},
                    2: {'description': 'two', 'revision_number': '222'}}
        for id in range(1, 4):
            fake_payload = self._fake_pf_payload_entry(None, id)

            with mock.patch.object(
                    self.pf_plugin, 'get_floatingip_port_forwardings',
                    return_value=[]):
                self._handle_notification_common(events.AFTER_DELETE,
                                                 fake_payload, fip_objs)
                calls = [mock.call(
                    mock.ANY, self.l3_plugin._nb_ovn, fake_payload.states[0])]
                self.handler.port_forwarding_deleted.assert_has_calls(calls)
                update_calls = [mock.call(
                    self.context, fake_payload.states[0].floatingip_id,
                    const.FLOATINGIP_STATUS_DOWN)]
                self.l3_plugin.update_floatingip_status.assert_has_calls(
                    update_calls)

    def test_maintenance_create_or_update(self):
        pf_objs = [self._fake_pf_obj()]
        fip_id = pf_objs[0].floatingip_id
        fake_fip_obj = {'floatingip_id': fip_id}
        fake_lb_names = ['lb1', 'lb2']
        self.handler.lb_names = mock.Mock(return_value=fake_lb_names)
        self.handler.port_forwarding_created = mock.Mock()
        self.l3_plugin.get_floatingip = mock.Mock(return_value=fake_fip_obj)
        with mock.patch.object(self._ovn_pf, '_get_pf_objs',
                               return_value=pf_objs) as mock_get_pf_objs:
            self._ovn_pf._maintenance_create_update(self.context, fip_id)
            self.l3_plugin._nb_ovn.transaction.assert_called_once_with(
                check_error=True)
            calls = [mock.call(lb_name, vip=None, if_exists=True)
                     for lb_name in fake_lb_names]
            self.l3_plugin._nb_ovn.lb_del.assert_has_calls(calls)
            calls = [mock.call(mock.ANY, self.l3_plugin._nb_ovn, pf_obj)
                     for pf_obj in pf_objs]
            self.handler.port_forwarding_created.assert_has_calls(calls)
            mock_get_pf_objs.assert_called_once_with(self.context, fip_id)
            self.l3_plugin.get_floatingip.assert_called_once_with(
                self.context, fip_id)
            self.fake_db_rev.assert_called_once_with(
                self.context, self.mock_pf_revs)

    def test_maintenance_delete(self):
        pf_objs = [self._fake_pf_obj()]
        fip_id = pf_objs[0].floatingip_id
        fake_fip_obj = {'floatingip_id': fip_id}
        fake_lb_names = ['lb1', 'lb2']
        self.handler.lb_names = mock.Mock(return_value=fake_lb_names)
        self.handler.port_forwarding_created = mock.Mock()
        self.l3_plugin.get_floatingip = mock.Mock(return_value=fake_fip_obj)
        with mock.patch.object(self._ovn_pf, '_get_pf_objs',
                               return_value=pf_objs) as mock_get_pf_objs:
            self._ovn_pf.maintenance_delete(self.context, fip_id)
            self.l3_plugin._nb_ovn.transaction.assert_called_once_with(
                check_error=True)
            calls = [mock.call(lb_name, vip=None, if_exists=True)
                     for lb_name in fake_lb_names]
            self.l3_plugin._nb_ovn.lb_del.assert_has_calls(calls)
            self.handler.port_forwarding_created.assert_not_called()
            mock_get_pf_objs.assert_not_called()
            self.l3_plugin.get_floatingip.assert_not_called()
            self.fake_db_rev.assert_not_called()

    @mock.patch.object(port_forwarding.LOG, 'info')
    def test_db_sync_create_or_update(self, m_info):
        pf_objs = [self._fake_pf_obj()]
        fip_id = pf_objs[0].floatingip_id
        fake_fip_obj = {'floatingip_id': fip_id, 'revision_number': 123456789}
        fake_lb_names = ['lb1', 'lb2']
        self.handler.lb_names = mock.Mock(return_value=fake_lb_names)
        self.handler.port_forwarding_created = mock.Mock()
        self.l3_plugin.get_floatingip = mock.Mock(return_value=fake_fip_obj)
        with mock.patch.object(self._ovn_pf, '_get_pf_objs',
                               return_value=pf_objs) as mock_get_pf_objs:
            self._ovn_pf.db_sync_create_or_update(
                self.context, fip_id, self.txn)
            info_args, _info_kwargs = m_info.call_args_list[0]
            self.assertIn('db_sync UPDATE entries', info_args[0])
            mock_get_pf_objs.assert_called_once_with(self.context, fip_id)
            calls = [mock.call(lb_name, vip=None, if_exists=True)
                     for lb_name in fake_lb_names]
            self.l3_plugin._nb_ovn.lb_del.assert_has_calls(calls)
            calls = [mock.call(mock.ANY, self.l3_plugin._nb_ovn, pf_obj)
                     for pf_obj in pf_objs]
            self.handler.port_forwarding_created.assert_has_calls(calls)
            self.l3_plugin.get_floatingip.assert_called_once_with(
                self.context, fip_id)
            self.fake_check_rev.assert_called_once_with(
                self.txn, self.l3_plugin._nb_ovn, fip_id, fake_fip_obj)

    @mock.patch.object(port_forwarding.LOG, 'info')
    def test_db_sync_delete(self, m_info):
        fip_id = 'fip_id'
        fake_lb_names = ['lb1', 'lb2', 'lb3', 'lb4', 'lb5']
        self.handler.lb_names = mock.Mock(return_value=fake_lb_names)
        self._ovn_pf.db_sync_delete(self.context, fip_id, self.txn)
        info_args, _info_kwargs = m_info.call_args_list[0]
        self.assertIn('db_sync DELETE entries', info_args[0])
        calls = [mock.call(lb_name, vip=None, if_exists=True)
                 for lb_name in fake_lb_names]
        self.l3_plugin._nb_ovn.lb_del.assert_has_calls(calls)



# Copyright 2016 Hewlett Packard Enterprise Development Company, LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as neutron_const
from neutron_lib.plugins import directory
from neutron_lib.services.trunk import constants
import testtools

from neutron.objects import trunk as trunk_objects
from neutron.services.trunk import drivers
from neutron.services.trunk import exceptions as trunk_exc
from neutron.services.trunk import plugin as trunk_plugin
from neutron.services.trunk import rules
from neutron.services.trunk.seg_types import validators
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron.tests.unit.services.trunk import fakes


def create_subport_dict(port_id):
    return {'segmentation_type': 'vlan',
            'segmentation_id': 123,
            'port_id': port_id}


def register_mock_callback(resource, event):
    callback = mock.Mock()
    registry.subscribe(callback, resource, event)
    return callback


class TrunkPluginTestCase(test_plugin.Ml2PluginV2TestCase):

    def setUp(self):
        super(TrunkPluginTestCase, self).setUp()
        self.drivers_patch = mock.patch.object(drivers, 'register').start()
        self.compat_patch = mock.patch.object(
            trunk_plugin.TrunkPlugin, 'check_compatibility').start()
        self.trunk_plugin = trunk_plugin.TrunkPlugin()
        self.trunk_plugin.add_segmentation_type('vlan', lambda x: True)

    def _create_test_trunk(self, port, subports=None):
        subports = subports if subports else []
        trunk = {'port_id': port['port']['id'],
                 'project_id': 'test_tenant',
                 'sub_ports': subports}
        response = (
            self.trunk_plugin.create_trunk(self.context, {'trunk': trunk}))
        return response

    def _get_trunk_obj(self, trunk_id):
        return trunk_objects.Trunk.get_object(self.context, id=trunk_id)

    def _get_subport_obj(self, port_id):
        subports = trunk_objects.SubPort.get_objects(
            self.context, port_id=port_id)
        return subports[0]

    def _test_delete_port_raise_in_use(self, parent_port, child_port, port_id,
                                       exception):
        subport = create_subport_dict(child_port['port']['id'])
        self._create_test_trunk(parent_port, [subport])
        core_plugin = directory.get_plugin()
        self.assertRaises(exception, core_plugin.delete_port,
                          self.context, port_id)

    def test_delete_port_raise_in_use_by_trunk(self):
        with self.port() as parent_port, self.port() as child_port:
            self._test_delete_port_raise_in_use(
                parent_port, child_port, parent_port['port']['id'],
                trunk_exc.PortInUseAsTrunkParent)

    def test_delete_port_raise_in_use_by_subport(self):
        with self.port() as parent_port, self.port() as child_port:
            self._test_delete_port_raise_in_use(
                parent_port, child_port, child_port['port']['id'],
                trunk_exc.PortInUseAsSubPort)

    def test_delete_trunk_raise_in_use(self):
        with self.port() as port:
            fakes.FakeDriverCanTrunkBoundPort.create()
            self.trunk_plugin = trunk_plugin.TrunkPlugin()
            directory.add_plugin('trunk', self.trunk_plugin)

            trunk = self._create_test_trunk(port)
            core_plugin = directory.get_plugin()
            port['port']['binding:host_id'] = 'host'
            core_plugin.update_port(self.context, port['port']['id'], port)

            trunk_port_validator = rules.TrunkPortValidator(trunk['port_id'])
            if not trunk_port_validator.can_be_trunked_or_untrunked(
                    self.context):
                self.assertRaises(trunk_exc.TrunkInUse,
                              self.trunk_plugin.delete_trunk,
                              self.context, trunk['id'])

    def _test_trunk_create_notify(self, event):
        with self.port() as parent_port:
            callback = register_mock_callback(resources.TRUNK, event)
            trunk = self._create_test_trunk(parent_port)
            trunk_obj = self._get_trunk_obj(trunk['id'])
            callback.assert_called_once_with(
                resources.TRUNK, event, self.trunk_plugin, payload=mock.ANY)
            payload = callback.mock_calls[0][2]['payload']
            self.assertEqual(self.context, payload.context)
            self.assertEqual(trunk_obj, payload.latest_state)
            self.assertEqual(trunk['id'], payload.resource_id)

    def test_create_trunk_notify_after_create(self):
        self._test_trunk_create_notify(events.AFTER_CREATE)

    def test_create_trunk_notify_precommit_create(self):
        self._test_trunk_create_notify(events.PRECOMMIT_CREATE)

    def _test_trunk_update_notify(self, event):
        with self.port() as parent_port:
            callback = register_mock_callback(resources.TRUNK, event)
            trunk = self._create_test_trunk(parent_port)
            orig_trunk_obj = self._get_trunk_obj(trunk['id'])
            trunk_req = {'trunk': {'name': 'foo'}}
            self.trunk_plugin.update_trunk(self.context, trunk['id'],
                                           trunk_req)
            trunk_obj = self._get_trunk_obj(trunk['id'])
            callback.assert_called_once_with(
                resources.TRUNK, event, self.trunk_plugin, payload=mock.ANY)
            payload = callback.mock_calls[0][2]['payload']
            self.assertEqual(self.context, payload.context)
            self.assertEqual(trunk_obj, payload.latest_state)
            self.assertEqual(trunk['id'], payload.resource_id)
            self.assertEqual(orig_trunk_obj, payload.states[0])

    def test_trunk_update_notify_after_update(self):
        self._test_trunk_update_notify(events.AFTER_UPDATE)

    def test_trunk_update_notify_precommit_update(self):
        self._test_trunk_update_notify(events.PRECOMMIT_UPDATE)

    def _test_trunk_delete_notify(self, event):
        with self.port() as parent_port:
            callback = register_mock_callback(resources.TRUNK, event)
            trunk = self._create_test_trunk(parent_port)
            trunk_obj = self._get_trunk_obj(trunk['id'])
            self.trunk_plugin.delete_trunk(self.context, trunk['id'])
            callback.assert_called_once_with(
                resources.TRUNK, event, self.trunk_plugin, payload=mock.ANY)
            payload = callback.mock_calls[0][2]['payload']
            self.assertEqual(self.context, payload.context)
            self.assertEqual(trunk_obj, payload.states[0])
            self.assertEqual(parent_port['port']['id'], payload.states[1].id)
            self.assertEqual(trunk['id'], payload.resource_id)

    def test_delete_trunk_notify_after_delete(self):
        self._test_trunk_delete_notify(events.AFTER_DELETE)

    def test_delete_trunk_notify_precommit_delete(self):
        self._test_trunk_delete_notify(events.PRECOMMIT_DELETE)

    def _test_subport_action_empty_list_no_notify(self, event, subport_method):
        with self.port() as parent_port:
            trunk = self._create_test_trunk(parent_port)
            callback = register_mock_callback(resources.SUBPORTS, event)
            subport_method(self.context, trunk['id'], {'sub_ports': []})
            callback.assert_not_called()

    def _test_add_subports_no_notification(self, event):
        self._test_subport_action_empty_list_no_notify(
            event, self.trunk_plugin.add_subports)

    def test_add_subports_notify_after_create_empty_list(self):
        self._test_add_subports_no_notification(events.AFTER_CREATE)

    def test_add_subports_notify_precommit_create_empty_list(self):
        self._test_add_subports_no_notification(events.PRECOMMIT_CREATE)

    def _test_remove_subports_no_notification(self, event):
        self._test_subport_action_empty_list_no_notify(
            event, self.trunk_plugin.remove_subports)

    def test_remove_subports_notify_after_delete_empty_list(self):
        self._test_remove_subports_no_notification(events.AFTER_DELETE)

    def test_remove_subports_notify_precommit_delete_empty_list(self):
        self._test_remove_subports_no_notification(events.PRECOMMIT_DELETE)

    def _test_add_subports_notify(self, event):
        with self.port() as parent_port, self.port() as child_port:
            trunk = self._create_test_trunk(parent_port)
            orig_trunk_obj = self._get_trunk_obj(trunk['id'])
            subport = create_subport_dict(child_port['port']['id'])
            callback = register_mock_callback(resources.SUBPORTS, event)
            self.trunk_plugin.add_subports(
                self.context, trunk['id'], {'sub_ports': [subport]})
            trunk_obj = self._get_trunk_obj(trunk['id'])
            subport_obj = self._get_subport_obj(subport['port_id'])
            callback.assert_called_once_with(
                resources.SUBPORTS, event, self.trunk_plugin, payload=mock.ANY)
            payload = callback.mock_calls[0][2]['payload']
            self.assertEqual(trunk['id'], payload.resource_id)
            self.assertEqual(trunk_obj, payload.latest_state)
            self.assertEqual(orig_trunk_obj, payload.states[0])
            self.assertEqual([subport_obj], payload.metadata['subports'])

    def test_add_subports_notify_after_create(self):
        self._test_add_subports_notify(events.AFTER_CREATE)

    def test_add_subports_notify_precommit_create(self):
        self._test_add_subports_notify(events.PRECOMMIT_CREATE)

    def _test_remove_subports_notify(self, event):
        with self.port() as parent_port, self.port() as child_port:
            subport = create_subport_dict(child_port['port']['id'])
            trunk = self._create_test_trunk(parent_port, [subport])
            orig_trunk_obj = self._get_trunk_obj(trunk['id'])
            callback = register_mock_callback(resources.SUBPORTS, event)
            subport_obj = self._get_subport_obj(subport['port_id'])
            self.trunk_plugin.remove_subports(
                self.context, trunk['id'], {'sub_ports': [subport]})
            trunk_obj = self._get_trunk_obj(trunk['id'])
            callback.assert_called_once_with(
                resources.SUBPORTS, event, self.trunk_plugin, payload=mock.ANY)
            payload = callback.mock_calls[0][2]['payload']
            self.assertEqual(trunk['id'], payload.resource_id)
            self.assertEqual(trunk_obj, payload.latest_state)
            self.assertEqual(orig_trunk_obj, payload.states[0])
            self.assertEqual([subport_obj], payload.metadata['subports'])

    def test_remove_subports_notify_after_delete(self):
        self._test_remove_subports_notify(events.AFTER_DELETE)

    def test_remove_subports_notify_precommit_delete(self):
        self._test_remove_subports_notify(events.PRECOMMIT_DELETE)

    def test_create_trunk_in_down_state(self):
        with self.port() as port:
            trunk = self._create_test_trunk(port)
            self.assertEqual(
                constants.TRUNK_DOWN_STATUS, trunk['status'])

    def test_add_subports_trunk_in_error_state_raises(self):
        with self.port() as port, self.port() as subport:
            trunk = self._create_test_trunk(port)
            trunk_obj = self._get_trunk_obj(trunk['id'])
            trunk_obj.status = constants.TRUNK_ERROR_STATUS
            trunk_obj.update()
            s = create_subport_dict(subport['port']['id'])
            self.assertRaises(trunk_exc.TrunkInErrorState,
                self.trunk_plugin.add_subports,
                self.context, trunk['id'], {'sub_ports': [s]})

    def test_add_subports_trunk_goes_to_down(self):
        with self.port() as port, self.port() as subport:
            trunk = self._create_test_trunk(port)
            trunk_obj = self._get_trunk_obj(trunk['id'])
            trunk_obj.status = constants.TRUNK_ACTIVE_STATUS
            trunk_obj.update()
            s = create_subport_dict(subport['port']['id'])
            trunk = self.trunk_plugin.add_subports(
                self.context, trunk['id'], {'sub_ports': [s]})
            self.assertEqual(constants.TRUNK_DOWN_STATUS, trunk['status'])

    def test_remove_subports_trunk_goes_to_down(self):
        with self.port() as port, self.port() as subport:
            s = create_subport_dict(subport['port']['id'])
            trunk = self._create_test_trunk(port, [s])
            trunk_obj = self._get_trunk_obj(trunk['id'])
            trunk_obj.status = constants.TRUNK_ACTIVE_STATUS
            trunk_obj.update()
            trunk = self.trunk_plugin.remove_subports(
                self.context, trunk['id'],
                {'sub_ports': [{'port_id': subport['port']['id']}]})
            self.assertEqual(constants.TRUNK_DOWN_STATUS, trunk['status'])

    def test__trigger_trunk_status_change_parent_port_status_down(self):
        callback = register_mock_callback(resources.TRUNK, events.AFTER_UPDATE)
        with self.port() as parent:
            parent['status'] = neutron_const.PORT_STATUS_DOWN
            original_port = {'status': neutron_const.PORT_STATUS_DOWN}
            _, _ = (
                self._test__trigger_trunk_status_change(
                    parent, original_port,
                    constants.TRUNK_DOWN_STATUS,
                    constants.TRUNK_DOWN_STATUS))
        callback.assert_not_called()

    def test__trigger_trunk_status_change_parent_port_status_up(self):
        callback = register_mock_callback(resources.TRUNK, events.AFTER_UPDATE)
        with self.port() as parent:
            parent['status'] = neutron_const.PORT_STATUS_ACTIVE
            original_port = {'status': neutron_const.PORT_STATUS_DOWN}
            _, _ = (
                self._test__trigger_trunk_status_change(
                    parent, original_port,
                    constants.TRUNK_DOWN_STATUS,
                    constants.TRUNK_ACTIVE_STATUS))
        callback.assert_called_once_with(
            resources.TRUNK, events.AFTER_UPDATE,
            self.trunk_plugin, payload=mock.ANY)

    def test__trigger_trunk_status_change_vif_type_changed_unbound(self):
        callback = register_mock_callback(resources.TRUNK, events.AFTER_UPDATE)
        with self.port() as parent:
            parent[portbindings.VIF_TYPE] = portbindings.VIF_TYPE_UNBOUND
            original_port = {portbindings.VIF_TYPE: 'fakeviftype'}
            original_trunk, current_trunk = (
                self._test__trigger_trunk_status_change(
                    parent, original_port,
                    constants.TRUNK_ACTIVE_STATUS,
                    constants.TRUNK_DOWN_STATUS))
        callback.assert_called_once_with(
            resources.TRUNK, events.AFTER_UPDATE,
            self.trunk_plugin, payload=mock.ANY)
        payload = callback.mock_calls[0][2]['payload']
        self.assertEqual(self.context, payload.context)
        self.assertEqual(current_trunk, payload.latest_state)
        self.assertEqual(original_trunk['id'], payload.resource_id)
        self.assertEqual(original_trunk, payload.states[0])

    def test__trigger_trunk_status_change_vif_type_unchanged(self):
        with self.port() as parent:
            parent[portbindings.VIF_TYPE] = 'fakeviftype'
            original_port = {portbindings.VIF_TYPE: 'fakeviftype'}
            self._test__trigger_trunk_status_change(
                parent, original_port, constants.TRUNK_ACTIVE_STATUS,
                constants.TRUNK_ACTIVE_STATUS)

    def test__trigger_trunk_status_change_vif_type_changed(self):
        with self.port() as parent:
            parent[portbindings.VIF_TYPE] = 'realviftype'
            original_port = {portbindings.VIF_TYPE: 'fakeviftype'}
            self._test__trigger_trunk_status_change(
                parent, original_port, constants.TRUNK_ACTIVE_STATUS,
                constants.TRUNK_ACTIVE_STATUS)

    def _test__trigger_trunk_status_change(self, new_parent,
                                           original_parent,
                                           initial_trunk_status,
                                           final_trunk_status):
        trunk = self._create_test_trunk(new_parent)
        trunk = self._get_trunk_obj(trunk['id'])
        trunk.update(status=initial_trunk_status)
        trunk_details = {'trunk_id': trunk.id}
        new_parent['trunk_details'] = trunk_details
        original_parent['trunk_details'] = trunk_details
        self.trunk_plugin._trigger_trunk_status_change(
            resources.PORT,
            events.AFTER_UPDATE,
            None,
            payload=events.DBEventPayload(
                self.context, states=(original_parent, new_parent)))
        current_trunk = self._get_trunk_obj(trunk.id)
        self.assertEqual(final_trunk_status, current_trunk.status)
        return trunk, current_trunk


class TrunkPluginCompatDriversTestCase(test_plugin.Ml2PluginV2TestCase):

    def setUp(self):
        super(TrunkPluginCompatDriversTestCase, self).setUp()
        mock.patch.object(drivers, 'register').start()

    def test_plugin_fails_to_start_no_loaded_drivers(self):
        with testtools.ExpectedException(
                trunk_exc.IncompatibleTrunkPluginConfiguration):
            trunk_plugin.TrunkPlugin()

    def test_plugins_fails_to_start_seg_type_validator_not_found(self):
        fakes.FakeDriver.create()
        with mock.patch.object(
                validators, 'get_validator', side_effect=KeyError), \
                testtools.ExpectedException(
                        trunk_exc.SegmentationTypeValidatorNotFound):
            trunk_plugin.TrunkPlugin()

    def test_plugins_fails_to_start_conflicting_seg_types(self):
        fakes.FakeDriver.create()
        fakes.FakeDriver2.create()
        with testtools.ExpectedException(
                trunk_exc.IncompatibleDriverSegmentationTypes):
            trunk_plugin.TrunkPlugin()

    def test_plugin_with_fake_driver(self):
        with mock.patch.object(validators, 'get_validator',
                               return_value={'foo_seg_types': mock.ANY}):
            fake_driver = fakes.FakeDriver.create()
            plugin = trunk_plugin.TrunkPlugin()
            self.assertTrue(fake_driver.is_loaded)
            self.assertEqual(set([]), plugin.supported_agent_types)
            self.assertEqual(set(['foo_intfs']), plugin.supported_interfaces)
            self.assertEqual([fake_driver], plugin.registered_drivers)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# TODO(flaviof): This file is a place holder. Everything here should move to
#                neutron-lib someday.

# String literals representing core resources.
# TODO(flaviof): move to neutron_lib/callbacks/resources.py
from ovsdbapp import constants as const

PORT_FORWARDING = 'port_forwarding'
PORT_FORWARDING_PLUGIN = 'port_forwarding_plugin'
PORT_FORWARDING_PREFIX = 'pf-floatingip'
LB_PROTOCOL_MAP = {'udp': const.PROTO_UDP, 'tcp': const.PROTO_TCP}



# Copyright (c) 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions
from oslo_upgradecheck import upgradecheck

from neutron.agent.linux import utils
from neutron.tests.functional import base


class StatusTest(base.BaseLoggingTestCase):

    def test_neutron_status_cli(self):
        """This test runs "neutron-status upgrade check" command and check if
        stdout contains header "Upgrade Check Results". It also checks if
        stderr contains only expected message.
        Example output from this CLI tool looks like:

        +----------------------------------------------------------------+
        | Upgrade Check Results                                          |
        +----------------------------------------------------------------+
        | Check: Worker counts configured                                |
        | Result: Warning                                                |
        | Details: The default number of workers has changed. Please see |
        |   release notes for the new values, but it is strongly         |
        |   encouraged for deployers to manually set the values for      |
        |   api_workers and rpc_workers.                                 |
        +----------------------------------------------------------------+

        Error codes which might be returned by this command:
        - Code.SUCCESS,
        - Code.WARNING,
        - Code.FAILURE
        are all accepted as we don't want to test here if there are any
        potential problems with upgrade or all is fine. This depends on
        deployment's configuration.
        """

        expected_result_title = "Upgrade Check Results"
        # NOTE(slaweq): it seems that ovsdbapp raises Exception() and prints
        # it's message to the stderr when it can't connect to the OVSDBs.
        # This upgrade check's test is just testing that tool is working fine
        # and don't really need to connect to the ovn databases so lets simply
        # expect that error message in the test
        expected_stderr = (
            'Unable to open stream to tcp:127.0.0.1:6641 to retrieve schema: '
            'Connection refused')
        try:
            stdout, stderr = utils.execute(
                cmd=["neutron-status", "upgrade", "check"],
                addl_env={'PYTHONWARNINGS': 'ignore'},
                extra_ok_codes=[upgradecheck.Code.SUCCESS,
                                upgradecheck.Code.WARNING,
                                upgradecheck.Code.FAILURE],
                return_stderr=True)
            self.assertEqual(
                expected_stderr,
                stderr.replace('\n', ''))
            self.assertTrue(expected_result_title in stdout)
        except exceptions.ProcessExecutionError as error:
            self.fail("neutron-status upgrade check command failed to run. "
                      "Error: %s" % error)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron._i18n import _


ip_opts = [
    cfg.BoolOpt('allsets',
                default=False,
                help=_('Destroy all IPsets.')),
    cfg.BoolOpt('force',
                default=False,
                help=_('Destroy IPsets even if there is an iptables '
                       'reference.')),
    cfg.StrOpt('prefix',
               default='N',  # ipset_manager.NET_PREFIX
               help=_('String prefix used to match IPset names.')),
]

netns_opts = [
    cfg.BoolOpt('force',
                default=False,
                help=_('Delete the namespace by removing all devices.')),
    cfg.StrOpt('agent-type',
               choices=['dhcp', 'l3'],
               help=_('Cleanup resources of a specific agent type only.')),
]

ovs_opts = [
    cfg.BoolOpt('ovs_all_ports',
                default=False,
                help=_('True to delete all ports on all the OpenvSwitch '
                       'bridges. False to delete ports created by '
                       'Neutron and Nova on the integration bridge.'))
]


def register_cmd_opts(opts, cfg=cfg.CONF):
    cfg.register_cli_opts(opts)



# Copyright 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import signal
from unittest import mock

from neutron_lib import constants
from neutron_lib import exceptions as lib_exceptions
from neutron_lib import fixture as lib_fixtures
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.agent.l3 import agent as l3_agent
from neutron.agent.l3 import router_info
from neutron.agent.linux import external_process as ep
from neutron.agent.linux import ip_lib
from neutron.agent.linux import iptables_manager
from neutron.agent.linux import utils as linux_utils
from neutron.agent.metadata import driver as metadata_driver
from neutron.common import metadata as comm_meta
from neutron.conf.agent import common as agent_config
from neutron.conf.agent.l3 import config as l3_config
from neutron.conf.agent.l3 import ha as ha_conf
from neutron.conf.agent.metadata import config as meta_conf
from neutron.tests import base
from neutron.tests.unit.agent.linux import test_utils

_uuid = uuidutils.generate_uuid


class FakeL3NATAgent(object):

    def __init__(self):
        self.conf = cfg.CONF


class TestMetadataDriverRules(base.BaseTestCase):

    def test_metadata_nat_rules(self):
        rules = ('PREROUTING', '-d 169.254.169.254/32 -i qr-+ '
                 '-p tcp -m tcp --dport 80 -j REDIRECT --to-ports 9697')
        self.assertEqual(
            [rules],
            metadata_driver.MetadataDriver.metadata_nat_rules(9697))

    def test_metadata_nat_rules_ipv6(self):
        rules = ('PREROUTING', '-d fe80::a9fe:a9fe/128 -i qr-+ '
                 '-p tcp -m tcp --dport 80 -j REDIRECT --to-ports 9697')
        self.assertEqual(
            [rules],
            metadata_driver.MetadataDriver.metadata_nat_rules(
                9697, metadata_address='fe80::a9fe:a9fe/128'))

    def test_metadata_filter_rules(self):
        rules = [('INPUT', '-m mark --mark 0x1/%s -j ACCEPT' %
                  constants.ROUTER_MARK_MASK),
                 ('INPUT', '-p tcp -m tcp --dport 9697 -j DROP')]
        self.assertEqual(
            rules,
            metadata_driver.MetadataDriver.metadata_filter_rules(9697, '0x1'))


class TestMetadataDriverProcess(base.BaseTestCase):

    EUNAME = 'neutron'
    EGNAME = 'neutron'
    METADATA_DEFAULT_IP = '169.254.169.254'
    METADATA_DEFAULT_IPV6 = 'fe80::a9fe:a9fe'
    METADATA_PORT = 8080
    METADATA_SOCKET = '/socket/path'
    PIDFILE = 'pidfile'
    RATE_LIMIT_CONFIG = {
        'base_window_duration': 10,
        'base_query_rate_limit': 5,
        'burst_window_duration': 1,
        'burst_query_rate_limit': 10,
    }

    def setUp(self):
        super(TestMetadataDriverProcess, self).setUp()
        mock.patch('eventlet.spawn').start()
        agent_config.register_interface_driver_opts_helper(cfg.CONF)
        cfg.CONF.set_override('interface_driver',
                              'neutron.agent.linux.interface.NullDriver')

        mock.patch('neutron.agent.l3.agent.L3PluginApi').start()
        mock.patch('neutron.agent.l3.ha.AgentMixin'
                   '._init_ha_conf_path').start()
        self.delete_if_exists = mock.patch.object(linux_utils,
                                                  'delete_if_exists').start()
        self.mock_get_process = mock.patch.object(
            metadata_driver.MetadataDriver,
            '_get_metadata_proxy_process_manager')

        l3_config.register_l3_agent_config_opts(l3_config.OPTS, cfg.CONF)
        ha_conf.register_l3_agent_ha_opts()
        meta_conf.register_meta_conf_opts(meta_conf.SHARED_OPTS, cfg.CONF)
        meta_conf.register_meta_conf_opts(
            meta_conf.METADATA_RATE_LIMITING_OPTS, cfg.CONF,
            group=meta_conf.RATE_LIMITING_GROUP)

    def test_after_router_updated_called_on_agent_process_update(self):
        with mock.patch.object(metadata_driver, 'after_router_updated') as f,\
                mock.patch('neutron.agent.l3.namespace_manager.'
                           'NamespaceManager.list_all', return_value={}),\
                mock.patch.object(router_info.RouterInfo, 'process'):
            agent = l3_agent.L3NATAgent('localhost')
            router_id = _uuid()
            router = {'id': router_id}
            ri = router_info.RouterInfo(mock.Mock(), router_id, router,
                                        agent.conf, mock.ANY)
            agent.router_info[router_id] = ri
            agent._process_updated_router(router)
            f.assert_called_once_with(
                'router', 'after_update', agent,
                payload=mock.ANY)

            payload = f.call_args_list[0][1]['payload']
            self.assertEqual(ri, payload.latest_state)
            self.assertEqual(router_id, payload.resource_id)

    def test_after_router_updated_should_not_call_add_metadata_rules(self):
        with mock.patch.object(iptables_manager.IptablesTable,
                               'add_rule') as f,\
                mock.patch.object(iptables_manager.IptablesManager,
                                  'apply'),\
                mock.patch.object(metadata_driver.MetadataDriver,
                                  'spawn_monitored_metadata_proxy'),\
                mock.patch('neutron.agent.l3.namespace_manager.'
                           'NamespaceManager.list_all', return_value={}),\
                mock.patch.object(router_info.RouterInfo, 'process'):
            agent = l3_agent.L3NATAgent('localhost')
            router_id = _uuid()
            router = {'id': router_id}
            ri = router_info.RouterInfo(mock.Mock(), router_id, router,
                                        agent.conf, mock.ANY)
            agent.router_info[router_id] = ri
            f.reset_mock()
            agent._process_updated_router(router)
            f.assert_not_called()

    def _test_spawn_metadata_proxy(self, dad_failed=False, rate_limited=False):
        router_id = _uuid()
        router_ns = 'qrouter-%s' % router_id
        service_name = 'haproxy'
        ip_class_path = 'neutron.agent.linux.ip_lib.IPWrapper'

        cfg.CONF.set_override('metadata_proxy_user', self.EUNAME)
        cfg.CONF.set_override('metadata_proxy_group', self.EGNAME)
        cfg.CONF.set_override('metadata_proxy_socket', self.METADATA_SOCKET)
        cfg.CONF.set_override('debug', True)

        with mock.patch(ip_class_path) as ip_mock,\
                mock.patch(
                    'neutron.agent.linux.external_process.'
                    'ProcessManager.get_pid_file_name',
                    return_value=self.PIDFILE),\
                mock.patch('pwd.getpwnam',
                           return_value=test_utils.FakeUser(self.EUNAME)),\
                mock.patch('grp.getgrnam',
                           return_value=test_utils.FakeGroup(self.EGNAME)),\
                mock.patch('os.makedirs'),\
                mock.patch('neutron.agent.l3.namespace_manager.'
                           'NamespaceManager.list_all', return_value={}),\
                mock.patch(
                    'neutron.agent.linux.ip_lib.'
                    'IpAddrCommand.wait_until_address_ready') as mock_wait,\
                mock.patch(
                    'neutron.agent.linux.ip_lib.'
                    'delete_ip_address') as mock_del:
            agent = l3_agent.L3NATAgent('localhost')
            agent.process_monitor = mock.Mock()
            cfg_file = os.path.join(
                metadata_driver.HaproxyConfigurator.get_config_path(
                    agent.conf.state_path),
                "%s.conf" % router_id)
            mock_open = self.useFixture(
                lib_fixtures.OpenFixture(cfg_file)).mock_open
            bind_v6_line = 'bind %s:%s interface %s' % (
                self.METADATA_DEFAULT_IPV6, self.METADATA_PORT, 'fake-if')
            if dad_failed:
                mock_wait.side_effect = ip_lib.DADFailed(
                    address=self.METADATA_DEFAULT_IPV6, reason='DAD failed')
                bind_v6_line = ''
            else:
                mock_wait.return_value = True
            agent.metadata_driver.spawn_monitored_metadata_proxy(
                agent.process_monitor,
                router_ns,
                self.METADATA_PORT,
                agent.conf,
                bind_address=self.METADATA_DEFAULT_IP,
                router_id=router_id,
                bind_address_v6=self.METADATA_DEFAULT_IPV6,
                bind_interface='fake-if')

            netns_execute_args = [
                service_name,
                '-f', cfg_file]

            log_tag = ("haproxy-" + metadata_driver.METADATA_SERVICE_NAME +
                       "-" + router_id)

            expected_params = {
                'user': self.EUNAME,
                'group': self.EGNAME,
                'host': self.METADATA_DEFAULT_IP,
                'port': self.METADATA_PORT,
                'unix_socket_path': self.METADATA_SOCKET,
                'res_type': 'Router',
                'res_id': router_id,
                'res_type_del': 'Network',
                'pidfile': self.PIDFILE,
                'log_level': 'debug',
                'log_tag': log_tag,
                'bind_v6_line': bind_v6_line}

            if dad_failed:
                mock_del.assert_called_once_with(self.METADATA_DEFAULT_IPV6,
                                                 'fake-if',
                                                 namespace=router_ns)
            else:
                mock_del.assert_not_called()

            if rate_limited:
                expected_params.update(self.RATE_LIMIT_CONFIG,
                                       stick_table_expire=10,
                                       ip_version='ip')
                expected_config_template = (
                    comm_meta.METADATA_HAPROXY_GLOBAL +
                    comm_meta.RATE_LIMITED_CONFIG_TEMPLATE +
                    metadata_driver._HEADER_CONFIG_TEMPLATE)
            else:
                expected_config_template = (
                    comm_meta.METADATA_HAPROXY_GLOBAL +
                    metadata_driver._UNLIMITED_CONFIG_TEMPLATE +
                    metadata_driver._HEADER_CONFIG_TEMPLATE)

            mock_open.assert_has_calls([
                mock.call(cfg_file, 'w'),
                mock.call().write(expected_config_template %
                                  expected_params)], any_order=True)

            env = {ep.PROCESS_TAG: service_name + '-' + router_id}
            ip_mock.assert_has_calls([
                mock.call(namespace=router_ns),
                mock.call().netns.execute(netns_execute_args, addl_env=env,
                                          run_as_root=True)
            ])

            agent.process_monitor.register.assert_called_once_with(
                router_id, metadata_driver.METADATA_SERVICE_NAME,
                mock.ANY)

            self.delete_if_exists.assert_called_once_with(
                mock.ANY, run_as_root=True)

    def test_spawn_metadata_proxy(self):
        self._test_spawn_metadata_proxy()

    def test_spawn_rate_limited_metadata_proxy(self):
        cfg.CONF.set_override('rate_limit_enabled', True,
                              group=meta_conf.RATE_LIMITING_GROUP)
        for k, v in self.RATE_LIMIT_CONFIG.items():
            cfg.CONF.set_override(k, v, group=meta_conf.RATE_LIMITING_GROUP)

        return self._test_spawn_metadata_proxy(rate_limited=True)

    def test_metadata_proxy_conf_parse_ip_versions(self):
        self.assertEqual(4, comm_meta.parse_ip_versions([4]))
        self.assertEqual(6, comm_meta.parse_ip_versions([6]))
        self.assertIsNone(comm_meta.parse_ip_versions([4, 6]))
        self.assertIsNone(comm_meta.parse_ip_versions([5, 6]))

    def test_spawn_metadata_proxy_dad_failed(self):
        self._test_spawn_metadata_proxy(dad_failed=True)

    @mock.patch.object(metadata_driver.LOG, 'error')
    def test_spawn_metadata_proxy_handles_process_exception(self, error_log):
        process_instance = mock.Mock(active=False)
        process_instance.enable.side_effect = (
            lib_exceptions.ProcessExecutionError('Something happened', -1))
        with mock.patch.object(metadata_driver.MetadataDriver,
                               '_get_metadata_proxy_process_manager',
                               return_value=process_instance):
            process_monitor = mock.Mock()
            network_id = 123456
            metadata_driver.MetadataDriver.spawn_monitored_metadata_proxy(
                process_monitor,
                'dummy_namespace',
                self.METADATA_PORT,
                cfg.CONF,
                network_id=network_id)
        error_log.assert_called_once()
        process_monitor.register.assert_not_called()
        self.assertNotIn(network_id, metadata_driver.MetadataDriver.monitors)

    def test_create_config_file_wrong_user(self):
        with mock.patch('pwd.getpwnam', side_effect=KeyError):
            config = metadata_driver.HaproxyConfigurator(_uuid(),
                                                         mock.ANY, mock.ANY,
                                                         mock.ANY, mock.ANY,
                                                         self.EUNAME,
                                                         self.EGNAME,
                                                         mock.ANY, mock.ANY,
                                                         mock.ANY)
            self.assertRaises(comm_meta.InvalidUserOrGroupException,
                              config.create_config_file)

    def test_create_config_file_wrong_group(self):
        with mock.patch('grp.getgrnam', side_effect=KeyError),\
                mock.patch('pwd.getpwnam',
                           return_value=test_utils.FakeUser(self.EUNAME)):
            config = metadata_driver.HaproxyConfigurator(_uuid(),
                                                         mock.ANY, mock.ANY,
                                                         mock.ANY, mock.ANY,
                                                         self.EUNAME,
                                                         self.EGNAME,
                                                         mock.ANY, mock.ANY,
                                                         mock.ANY)
            self.assertRaises(comm_meta.InvalidUserOrGroupException,
                              config.create_config_file)

    def test_destroy_monitored_metadata_proxy(self):
        mproxy_process = mock.Mock(active=False)
        mock_get_process = self.mock_get_process.start()
        mock_get_process.return_value = mproxy_process
        driver = metadata_driver.MetadataDriver(FakeL3NATAgent())
        driver.destroy_monitored_metadata_proxy(mock.Mock(), 'uuid', 'conf',
                                                'ns_name')
        mproxy_process.disable.assert_called_once_with(
            sig=str(int(signal.SIGTERM)))
        self.delete_if_exists.assert_called_once_with(
            mock.ANY, run_as_root=True)

    def test_destroy_monitored_metadata_proxy_force(self):
        mproxy_process = mock.Mock(active=True)
        mock_get_process = self.mock_get_process.start()
        mock_get_process.return_value = mproxy_process
        driver = metadata_driver.MetadataDriver(FakeL3NATAgent())
        with mock.patch.object(metadata_driver, 'SIGTERM_TIMEOUT', 0):
            driver.destroy_monitored_metadata_proxy(mock.Mock(), 'uuid',
                                                    'conf', 'ns_name')
        mproxy_process.disable.assert_has_calls([
            mock.call(sig=str(int(signal.SIGTERM))),
            mock.call(sig=str(int(signal.SIGKILL)))])
        self.delete_if_exists.assert_called_once_with(
            mock.ANY, run_as_root=True)



# Copyright (c) 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from oslo_config import cfg

from neutron.conf.agent import ovs_conf as agent_ovs_conf
from neutron.conf.plugins.ml2.drivers import ovs_conf as ml2_ovs_conf
from neutron.privileged.agent.ovsdb.native import helpers as priv_helpers


agent_ovs_conf.register_ovs_agent_opts(cfg.CONF)
ml2_ovs_conf.register_ovs_opts(cfg=cfg.CONF)

enable_connection_uri = functools.partial(
    priv_helpers.enable_connection_uri,
    log_fail_as_error=False, check_exit_code=False,
    timeout=cfg.CONF.OVS.ovsdb_timeout,
    inactivity_probe=cfg.CONF.OVS.of_inactivity_probe * 1000)



# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from neutron_lib.agent import topics
from neutron_lib.api.definitions import metering as metering_apidef
from neutron_lib import context
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from neutron_lib.tests import tools
from neutron_lib.utils import net as net_utils

from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from neutron.api.rpc.agentnotifiers import metering_rpc_agent_api
from neutron.db.metering import metering_rpc
from neutron.extensions import l3 as ext_l3
from neutron.extensions import metering as ext_metering
from neutron.objects import agent as agent_obj
from neutron.tests.common import helpers
from neutron.tests.unit.db.metering import test_metering_db
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit.extensions import test_l3


_uuid = uuidutils.generate_uuid

METERING_SERVICE_PLUGIN_KLASS = (
    "neutron.services.metering."
    "metering_plugin.MeteringPlugin"
)


class MeteringTestExtensionManager(object):

    def get_resources(self):
        l3_res = ext_l3.L3.get_resources()
        metering_res = ext_metering.Metering.get_resources()

        return l3_res + metering_res

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class TestMeteringPlugin(test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
                         test_l3.L3NatTestCaseMixin,
                         test_metering_db.MeteringPluginDbTestCaseMixin):

    resource_prefix_map = dict(
        (k.replace('_', '-'), "/metering")
        for k in metering_apidef.RESOURCE_ATTRIBUTE_MAP.keys()
    )

    def setUp(self):
        plugin = 'neutron.tests.unit.extensions.test_l3.TestL3NatIntPlugin'
        service_plugins = {'metering_plugin_name':
                           METERING_SERVICE_PLUGIN_KLASS}
        ext_mgr = MeteringTestExtensionManager()
        super(TestMeteringPlugin, self).setUp(plugin=plugin, ext_mgr=ext_mgr,
                                              service_plugins=service_plugins)

        self.uuid = '654f6b9d-0f36-4ae5-bd1b-01616794ca60'

        uuid = 'oslo_utils.uuidutils.generate_uuid'
        self.uuid_patch = mock.patch(uuid, return_value=self.uuid)
        self.mock_uuid = self.uuid_patch.start()

        self.ctx = context.Context('', self._tenant_id).elevated()

        self.topic = topics.METERING_AGENT

        add = ('neutron.api.rpc.agentnotifiers.' +
               'metering_rpc_agent_api.MeteringAgentNotifyAPI' +
               '.add_metering_label')
        self.add_patch = mock.patch(add)
        self.mock_add = self.add_patch.start()

        remove = ('neutron.api.rpc.agentnotifiers.' +
                  'metering_rpc_agent_api.MeteringAgentNotifyAPI' +
                  '.remove_metering_label')
        self.remove_patch = mock.patch(remove)
        self.mock_remove = self.remove_patch.start()

        update = ('neutron.api.rpc.agentnotifiers.' +
                  'metering_rpc_agent_api.MeteringAgentNotifyAPI' +
                  '.update_metering_label_rules')
        self.update_patch = mock.patch(update)
        self.mock_update = self.update_patch.start()

        add_rule = ('neutron.api.rpc.agentnotifiers.' +
                    'metering_rpc_agent_api.MeteringAgentNotifyAPI' +
                    '.add_metering_label_rule')
        self.add_rule_patch = mock.patch(add_rule)
        self.mock_add_rule = self.add_rule_patch.start()

        remove_rule = ('neutron.api.rpc.agentnotifiers.' +
                       'metering_rpc_agent_api.MeteringAgentNotifyAPI' +
                       '.remove_metering_label_rule')
        self.remove_rule_patch = mock.patch(remove_rule)
        self.mock_remove_rule = self.remove_rule_patch.start()

    def test_routers_updated_on_host_rpc_call(self):
        router_test = {
            'id': 'xyz',
            'name': 'testrouter'}
        notify_host = ('neutron.api.rpc.agentnotifiers.' +
                       'metering_rpc_agent_api.MeteringAgentNotifyAPI' +
                       '._notification_host')
        self.notify_patch = mock.patch(notify_host)
        self.mock_notify_host = self.notify_patch.start()
        metering_rpc_handle = metering_rpc_agent_api.MeteringAgentNotifyAPI()
        metering_rpc_handle.routers_updated_on_host(
            self.ctx,
            [router_test['id']],
            'test_host')
        self.mock_notify_host.assert_called_with(self.ctx, 'routers_updated',
                                                 'test_host', routers=['xyz'])

    def test_add_metering_label_rpc_call(self):
        second_uuid = 'e27fe2df-376e-4ac7-ae13-92f050a21f84'
        expected = [{'status': 'ACTIVE',
                     'name': 'router1',
                     'gw_port_id': None,
                     'admin_state_up': True,
                     'distributed': False,
                     'project_id': self._tenant_id,
                     '_metering_labels': [
                         {'rules': [],
                          'id': self.uuid, 'shared': False,
                          'name': 'label'}],
                     'id': self.uuid}]

        tenant_id_2 = '8a268a58-1610-4890-87e0-07abb8231206'
        self.mock_uuid.return_value = second_uuid
        with self.router(name='router2', tenant_id=tenant_id_2,
                         set_context=True):
            self.mock_uuid.return_value = self.uuid
            with self.router(name='router1'):
                with self.metering_label():
                    self.mock_add.assert_called_with(mock.ANY, expected)

    def test_add_metering_label_shared_rpc_call(self):
        second_uuid = 'e27fe2df-376e-4ac7-ae13-92f050a21f84'
        expected = [{'status': 'ACTIVE',
                     'name': 'router1',
                     'gw_port_id': None,
                     'admin_state_up': True,
                     'distributed': False,
                     'project_id': self._tenant_id,
                     '_metering_labels': [
                         {'rules': [],
                          'id': self.uuid, 'shared': False,
                          'name': 'label'},
                         {'rules': [],
                          'id': second_uuid, 'shared': True,
                          'name': 'label'}],
                     'id': self.uuid}]

        tenant_id_2 = '8a268a58-1610-4890-87e0-07abb8231206'
        with self.router(name='router1', shared=True):
            with self.metering_label():
                self.mock_uuid.return_value = second_uuid
                with self.metering_label(tenant_id=tenant_id_2, shared=True):
                    self.mock_add.assert_called_with(mock.ANY, expected)

    def test_remove_metering_label_rpc_call(self):
        expected = [{'status': 'ACTIVE',
                     'name': 'router1',
                     'gw_port_id': None,
                     'admin_state_up': True,
                     'distributed': False,
                     'project_id': self._tenant_id,
                     '_metering_labels': [
                         {'rules': [],
                          'id': self.uuid, 'shared': False,
                          'name': 'label'}],
                     'id': self.uuid}]

        with self.router():
            with self.metering_label() as label:
                self.mock_add.assert_called_with(mock.ANY, expected)
                self._delete('metering-labels',
                             label['metering_label']['id'],
                             as_admin=True)
            self.mock_remove.assert_called_with(mock.ANY, expected)

    def test_remove_one_metering_label_rpc_call(self):
        second_uuid = 'e27fe2df-376e-4ac7-ae13-92f050a21f84'
        expected_add = [{'status': 'ACTIVE',
                         'name': 'router1',
                         'gw_port_id': None,
                         'admin_state_up': True,
                         'distributed': False,
                         'project_id': self._tenant_id,
                         '_metering_labels': [
                             {'rules': [],
                              'id': self.uuid, 'shared': False,
                              'name': 'label'},
                             {'rules': [],
                              'id': second_uuid, 'shared': False,
                              'name': 'label'}],
                         'id': self.uuid}]
        expected_remove = [{'status': 'ACTIVE',
                            'name': 'router1',
                            'gw_port_id': None,
                            'admin_state_up': True,
                            'distributed': False,
                            'project_id': self._tenant_id,
                            '_metering_labels': [
                                {'rules': [],
                                 'id': second_uuid, 'shared': False,
                                 'name': 'label'}],
                            'id': self.uuid}]

        with self.router():
            with self.metering_label():
                self.mock_uuid.return_value = second_uuid
                with self.metering_label() as label:
                    self.mock_add.assert_called_with(mock.ANY, expected_add)
                    self._delete('metering-labels',
                                 label['metering_label']['id'],
                                 as_admin=True)
                self.mock_remove.assert_called_with(mock.ANY, expected_remove)

    def test_add_and_remove_metering_label_rule_rpc_call(self):
        second_uuid = 'e27fe2df-376e-4ac7-ae13-92f050a21f84'
        expected_add = [{'status': 'ACTIVE',
                         'name': 'router1',
                         'gw_port_id': None,
                         'admin_state_up': True,
                         'distributed': False,
                         'project_id': self._tenant_id,
                         '_metering_labels': [
                             {'rule': {
                                 'remote_ip_prefix':
                                     net_utils.AuthenticIPNetwork(
                                         '10.0.0.0/24'),
                                 'destination_ip_prefix': None,
                                 'source_ip_prefix': None,
                                 'direction': 'ingress',
                                 'metering_label_id': self.uuid,
                                 'excluded': False,
                                 'id': second_uuid},
                             'id': self.uuid}],
                         'id': self.uuid}]

        expected_del = [{'status': 'ACTIVE',
                         'name': 'router1',
                         'gw_port_id': None,
                         'admin_state_up': True,
                         'distributed': False,
                         'project_id': self._tenant_id,
                         '_metering_labels': [
                             {'rule': {
                                 'remote_ip_prefix':
                                     net_utils.AuthenticIPNetwork(
                                         '10.0.0.0/24'),
                                 'destination_ip_prefix': None,
                                 'source_ip_prefix': None,
                                 'direction': 'ingress',
                                 'metering_label_id': self.uuid,
                                 'excluded': False,
                                 'id': second_uuid},
                                 'id': self.uuid}],
                         'id': self.uuid}]

        remote_ip_prefix = {'remote_ip_prefix': '10.0.0.0/24'}
        with self.router():
            with self.metering_label() as label:
                la = label['metering_label']
                self.mock_uuid.return_value = second_uuid
                with self.metering_label_rule(la['id'], **remote_ip_prefix):
                    self.mock_add_rule.assert_called_with(mock.ANY,
                                                          expected_add)
                    self._delete('metering-label-rules', second_uuid,
                                 as_admin=True)
                self.mock_remove_rule.assert_called_with(mock.ANY,
                                                         expected_del)

    def test_add_and_remove_metering_label_rule_source_ip_only(self):
        second_uuid = 'e27fe2df-376e-4ac7-ae13-92f050a21f84'
        expected_add = [{'status': 'ACTIVE',
                         'name': 'router1',
                         'gw_port_id': None,
                         'admin_state_up': True,
                         'distributed': False,
                         'project_id': self._tenant_id,
                         '_metering_labels': [
                             {'rule': {
                                 'source_ip_prefix':
                                     net_utils.AuthenticIPNetwork(
                                         '10.0.0.0/24'),
                                 'destination_ip_prefix': None,
                                 'remote_ip_prefix': None,
                                 'direction': 'ingress',
                                 'metering_label_id': self.uuid,
                                 'excluded': False,
                                 'id': second_uuid},
                              'id': self.uuid}],
                         'id': self.uuid}]

        expected_del = [{'status': 'ACTIVE',
                         'name': 'router1',
                         'gw_port_id': None,
                         'admin_state_up': True,
                         'distributed': False,
                         'project_id': self._tenant_id,
                         '_metering_labels': [
                             {'rule': {
                                 'source_ip_prefix':
                                     net_utils.AuthenticIPNetwork(
                                         '10.0.0.0/24'),
                                 'destination_ip_prefix': None,
                                 'remote_ip_prefix': None,
                                 'direction': 'ingress',
                                 'metering_label_id': self.uuid,
                                 'excluded': False,
                                 'id': second_uuid},
                              'id': self.uuid}],
                         'id': self.uuid}]

        source_ip_prefix = {'source_ip_prefix': '10.0.0.0/24'}
        with self.router():
            with self.metering_label() as label:
                la = label['metering_label']
                self.mock_uuid.return_value = second_uuid
                with self.metering_label_rule(la['id'],
                                              **source_ip_prefix):
                    self.mock_add_rule.assert_called_with(mock.ANY,
                                                          expected_add)
                    self._delete('metering-label-rules', second_uuid,
                                 as_admin=True)
                self.mock_remove_rule.assert_called_with(mock.ANY,
                                                         expected_del)

    def test_add_and_remove_metering_label_rule_dest_ip_only(self):
        second_uuid = 'e27fe2df-376e-4ac7-ae13-92f050a21f84'
        expected_add = [{'status': 'ACTIVE',
                         'name': 'router1',
                         'gw_port_id': None,
                         'admin_state_up': True,
                         'distributed': False,
                         'project_id': self._tenant_id,
                         '_metering_labels': [
                             {'rule': {
                                 'destination_ip_prefix':
                                     net_utils.AuthenticIPNetwork(
                                         '10.0.0.0/24'),
                                 'source_ip_prefix': None,
                                 'remote_ip_prefix': None,
                                 'direction': 'ingress',
                                 'metering_label_id': self.uuid,
                                 'excluded': False,
                                 'id': second_uuid},
                              'id': self.uuid}],
                         'id': self.uuid}]

        expected_del = [{'status': 'ACTIVE',
                         'name': 'router1',
                         'gw_port_id': None,
                         'admin_state_up': True,
                         'distributed': False,
                         'project_id': self._tenant_id,
                         '_metering_labels': [
                             {'rule': {
                                 'destination_ip_prefix':
                                     net_utils.AuthenticIPNetwork(
                                         '10.0.0.0/24'),
                                 'source_ip_prefix': None,
                                 'remote_ip_prefix': None,
                                 'direction': 'ingress',
                                 'metering_label_id': self.uuid,
                                 'excluded': False,
                                 'id': second_uuid},
                              'id': self.uuid}],
                         'id': self.uuid}]

        source_ip_prefix = {'destination_ip_prefix': '10.0.0.0/24'}
        with self.router():
            with self.metering_label() as label:
                la = label['metering_label']
                self.mock_uuid.return_value = second_uuid
                with self.metering_label_rule(la['id'],
                                              **source_ip_prefix):
                    self.mock_add_rule.assert_called_with(mock.ANY,
                                                          expected_add)
                    self._delete('metering-label-rules', second_uuid,
                                 as_admin=True)
                self.mock_remove_rule.assert_called_with(mock.ANY,
                                                         expected_del)

    def test_add_and_remove_metering_label_rule_src_and_dest_ip_only(self):
        second_uuid = 'e27fe2df-376e-4ac7-ae13-92f050a21f84'
        expected_add = [{'status': 'ACTIVE',
                         'name': 'router1',
                         'gw_port_id': None,
                         'admin_state_up': True,
                         'distributed': False,
                         'project_id': self._tenant_id,
                         '_metering_labels': [
                             {'rule': {
                                 'destination_ip_prefix':
                                     net_utils.AuthenticIPNetwork('0.0.0.0/0'),
                                 'source_ip_prefix':
                                     net_utils.AuthenticIPNetwork(
                                         '10.0.0.0/24'),
                                 'remote_ip_prefix': None,
                                 'direction': 'ingress',
                                 'metering_label_id': self.uuid,
                                 'excluded': False,
                                 'id': second_uuid},
                              'id': self.uuid}],
                         'id': self.uuid}]

        expected_del = [{'status': 'ACTIVE',
                         'name': 'router1',
                         'gw_port_id': None,
                         'admin_state_up': True,
                         'distributed': False,
                         'project_id': self._tenant_id,
                         '_metering_labels': [
                             {'rule': {
                                 'destination_ip_prefix':
                                     net_utils.AuthenticIPNetwork('0.0.0.0/0'),
                                 'source_ip_prefix':
                                     net_utils.AuthenticIPNetwork(
                                         '10.0.0.0/24'),
                                 'remote_ip_prefix': None,
                                 'direction': 'ingress',
                                 'metering_label_id': self.uuid,
                                 'excluded': False,
                                 'id': second_uuid},
                              'id': self.uuid}],
                         'id': self.uuid}]

        ip_prefixes = {'source_ip_prefix': '10.0.0.0/24',
                       'destination_ip_prefix': '0.0.0.0/0'}
        with self.router():
            with self.metering_label() as label:
                la = label['metering_label']
                self.mock_uuid.return_value = second_uuid
                with self.metering_label_rule(la['id'],
                                              **ip_prefixes):
                    self.mock_add_rule.assert_called_with(mock.ANY,
                                                          expected_add)
                    self._delete('metering-label-rules', second_uuid,
                                 as_admin=True)
                self.mock_remove_rule.assert_called_with(mock.ANY,
                                                         expected_del)

    def test_add_and_remove_metering_label_rule_src_and_remote_ip(self):
        with self.router():
            with self.metering_label() as label:
                la = label['metering_label']

                res = self._create_metering_label_rule(
                    self.fmt, la['id'], 'ingress', False,
                    remote_ip_prefix='0.0.0.0/0',
                    source_ip_prefix='10.0.0.0/24')

                expected_error_code = 500
                self.assertEqual(expected_error_code, res.status_int)

                expected_error_message = "Cannot use 'remote-ip-prefix' in " \
                                         "conjunction with " \
                                         "'source-ip-prefix' or " \
                                         "'destination-ip-prefix'."

                self.assertEqual(
                    expected_error_message, jsonutils.loads(res.body)[
                        "NeutronError"]["message"])

    def test_add_and_remove_metering_label_rule_dest_and_remote_ip(self):
        with self.router():
            with self.metering_label() as label:
                la = label['metering_label']

                res = self._create_metering_label_rule(
                    self.fmt, la['id'], 'ingress', False,
                    remote_ip_prefix='0.0.0.0/0',
                    destination_ip_prefix='8.8.8.8/32')

                expected_error_code = 500
                self.assertEqual(expected_error_code, res.status_int)

                expected_error_message = "Cannot use 'remote-ip-prefix' in " \
                                         "conjunction with " \
                                         "'source-ip-prefix' or " \
                                         "'destination-ip-prefix'."

                self.assertEqual(
                    expected_error_message, jsonutils.loads(res.body)[
                        "NeutronError"]["message"])

    def test_add_and_remove_metering_label_rule_no_ip_prefix_entered(self):
        with self.router():
            with self.metering_label() as label:
                la = label['metering_label']

                res = self._create_metering_label_rule(
                    self.fmt, la['id'], 'ingress', False)

                expected_error_code = 500
                self.assertEqual(expected_error_code, res.status_int)

                expected_error_message = "You must define at least one of " \
                                         "the following parameters " \
                                         "'remote_ip_prefix', or " \
                                         "'source_ip_prefix' or " \
                                         "'destination_ip_prefix'."

                self.assertEqual(
                    expected_error_message, jsonutils.loads(res.body)[
                        "NeutronError"]["message"])

    def test_delete_metering_label_does_not_clear_router_tenant_id(self):
        tenant_id = '654f6b9d-0f36-4ae5-bd1b-01616794ca60'
        # TODO(ralonsoh): to investigate why the context in [1] has some value
        # in session.transaction._connections, while during a normal operation,
        # the ._connections value is empty.
        # [1]https://github.com/openstack/neutron/blob/
        # 1b9e9a6c2ccf7f9bc06429f53e5126f356ae3d4a/neutron/api/v2/base.py#L563
        self.ctx.GUARD_TRANSACTION = False
        with self.metering_label(tenant_id=tenant_id) as metering_label:
            with self.router(tenant_id=tenant_id) as r:
                router = self._show('routers', r['router']['id'],
                                    tenant_id=tenant_id)
                self.assertEqual(tenant_id, router['router']['tenant_id'])
                metering_label_id = metering_label['metering_label']['id']
                self._delete('metering-labels', metering_label_id, 204,
                             as_admin=True)
                router = self._show('routers', r['router']['id'],
                                    tenant_id=tenant_id)
                self.assertEqual(tenant_id, router['router']['tenant_id'])


class TestMeteringPluginL3AgentScheduler(
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
        test_l3.L3NatTestCaseMixin,
        test_metering_db.MeteringPluginDbTestCaseMixin):

    resource_prefix_map = dict(
        (k.replace('_', '-'), "/metering")
        for k in metering_apidef.RESOURCE_ATTRIBUTE_MAP.keys()
    )

    def setUp(self, plugin_str=None, service_plugins=None, scheduler=None):
        if not plugin_str:
            plugin_str = ('neutron.tests.unit.extensions.test_l3.'
                          'TestL3NatIntAgentSchedulingPlugin')

        if not service_plugins:
            service_plugins = {'metering_plugin_name':
                               METERING_SERVICE_PLUGIN_KLASS}

        if not scheduler:
            scheduler = plugin_str

        ext_mgr = MeteringTestExtensionManager()
        super(TestMeteringPluginL3AgentScheduler,
              self).setUp(plugin=plugin_str, ext_mgr=ext_mgr,
                          service_plugins=service_plugins)

        self.uuid = '654f6b9d-0f36-4ae5-bd1b-01616794ca60'

        uuid = 'oslo_utils.uuidutils.generate_uuid'
        self.uuid_patch = mock.patch(uuid, return_value=self.uuid)
        self.mock_uuid = self.uuid_patch.start()

        self.ctx = context.Context('', self._tenant_id).elevated()

        self.l3routers_patch = mock.patch(scheduler +
                                          '.get_l3_agents_hosting_routers')
        self.l3routers_mock = self.l3routers_patch.start()

        self.topic = topics.METERING_AGENT

        add = ('neutron.api.rpc.agentnotifiers.' +
               'metering_rpc_agent_api.MeteringAgentNotifyAPI' +
               '.add_metering_label')
        self.add_patch = mock.patch(add)
        self.mock_add = self.add_patch.start()

        remove = ('neutron.api.rpc.agentnotifiers.' +
                  'metering_rpc_agent_api.MeteringAgentNotifyAPI' +
                  '.remove_metering_label')
        self.remove_patch = mock.patch(remove)
        self.mock_remove = self.remove_patch.start()

    def test_add_metering_label_rpc_call(self):
        second_uuid = 'e27fe2df-376e-4ac7-ae13-92f050a21f84'
        expected = [{'status': 'ACTIVE',
                     'name': 'router1',
                     'gw_port_id': None,
                     'admin_state_up': True,
                     'distributed': False,
                     'project_id': self._tenant_id,
                     '_metering_labels': [
                         {'rules': [],
                          'id': second_uuid, 'shared': False,
                          'name': 'label'}],
                     'id': self.uuid},
                    {'status': 'ACTIVE',
                     'name': 'router2',
                     'gw_port_id': None,
                     'admin_state_up': True,
                     'distributed': False,
                     'project_id': self._tenant_id,
                     '_metering_labels': [
                         {'rules': [],
                          'id': second_uuid, 'shared': False,
                          'name': 'label'}],
                     'id': second_uuid}]

        # bind each router to a specific agent
        agent1 = agent_obj.Agent(mock.ANY, host='agent1')
        agent2 = agent_obj.Agent(mock.ANY, host='agent2')

        agents = {self.uuid: agent1,
                  second_uuid: agent2}

        def side_effect(context, routers, admin_state_up, active):
            return [agents[routers[0]]]

        self.l3routers_mock.side_effect = side_effect

        with self.router(name='router1'):
            self.mock_uuid.return_value = second_uuid
            with self.router(name='router2'):
                with self.metering_label():
                    self.mock_add.assert_called_with(
                        mock.ANY, tools.UnorderedList(expected))


class TestMeteringPluginL3AgentSchedulerServicePlugin(
        TestMeteringPluginL3AgentScheduler):

    """Unit tests for the case where separate service plugin
    implements L3 routing.
    """

    def setUp(self):
        l3_plugin = ('neutron.tests.unit.extensions.test_l3.'
                     'TestL3NatAgentSchedulingServicePlugin')
        service_plugins = {'metering_plugin_name':
                           METERING_SERVICE_PLUGIN_KLASS,
                           'l3_plugin_name': l3_plugin}

        plugin_str = ('neutron.tests.unit.extensions.test_l3.'
                      'TestNoL3NatPlugin')

        super(TestMeteringPluginL3AgentSchedulerServicePlugin, self).setUp(
            plugin_str=plugin_str, service_plugins=service_plugins,
            scheduler=l3_plugin)


class TestMeteringPluginRpcFromL3Agent(
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
        test_l3.L3NatTestCaseMixin,
        test_metering_db.MeteringPluginDbTestCaseMixin):

    resource_prefix_map = dict(
        (k.replace('_', '-'), "/metering")
        for k in metering_apidef.RESOURCE_ATTRIBUTE_MAP
    )

    def setUp(self):
        service_plugins = {'metering_plugin_name':
                           METERING_SERVICE_PLUGIN_KLASS}

        plugin = ('neutron.tests.unit.extensions.test_l3.'
                  'TestL3NatIntAgentSchedulingPlugin')

        ext_mgr = MeteringTestExtensionManager()
        super(TestMeteringPluginRpcFromL3Agent,
              self).setUp(plugin=plugin, service_plugins=service_plugins,
                          ext_mgr=ext_mgr)

        self.meter_plugin = directory.get_plugin(constants.METERING)

        self.tenant_id_1 = 'tenant_id_1'
        self.tenant_id_2 = 'tenant_id_2'

        self.adminContext = context.get_admin_context()
        helpers.register_l3_agent(host='agent1')

    def test_get_sync_data_metering(self):
        with self.subnet() as subnet:
            s = subnet['subnet']
            self._set_net_external(s['network_id'])
            with self.router(name='router1', subnet=subnet) as router:
                r = router['router']
                self._add_external_gateway_to_router(r['id'], s['network_id'])
                with self.metering_label(tenant_id=r['tenant_id']):
                    callbacks = metering_rpc.MeteringRpcCallbacks(
                        self.meter_plugin)
                    data = callbacks.get_sync_data_metering(self.adminContext,
                                                            host='agent1')
                    self.assertEqual('router1', data[0]['name'])

                    helpers.register_l3_agent(host='agent2')
                    data = callbacks.get_sync_data_metering(self.adminContext,
                                                            host='agent2')
                    self.assertFalse(data)

                self._remove_external_gateway_from_router(
                    r['id'], s['network_id'])

    def test_get_sync_data_metering_shared(self):
        with self.router(name='router1', tenant_id=self.tenant_id_1):
            with self.router(name='router2', tenant_id=self.tenant_id_2):
                with self.metering_label(shared=True):
                    callbacks = metering_rpc.MeteringRpcCallbacks(
                        self.meter_plugin)
                    data = callbacks.get_sync_data_metering(self.adminContext)

                    routers = [router['name'] for router in data]

                    self.assertIn('router1', routers)
                    self.assertIn('router2', routers)

    def test_get_sync_data_metering_not_shared(self):
        with self.router(name='router1', tenant_id=self.tenant_id_1):
            with self.router(name='router2', tenant_id=self.tenant_id_2):
                with self.metering_label():
                    callbacks = metering_rpc.MeteringRpcCallbacks(
                        self.meter_plugin)
                    data = callbacks.get_sync_data_metering(self.adminContext)

                    routers = [router['name'] for router in data]

                    self.assertEqual([], routers)

    def test_get_sync_data_metering_with_unscheduled_router(self):
        with self.subnet() as subnet:
            s = subnet['subnet']
            self._set_net_external(s['network_id'])
            with self.router(name='router1') as router1:
                self._add_external_gateway_to_router(
                    router1['router']['id'], s['network_id'])
                with self.router(name='router2'):
                    with self.metering_label():
                        callbacks = metering_rpc.MeteringRpcCallbacks(
                            self.meter_plugin)
                        data = callbacks.get_sync_data_metering(
                            self.adminContext, host='agent1')
                        self.assertEqual(
                            set(['router1']), set([r['name'] for r in data]))

                self._remove_external_gateway_from_router(
                    router1['router']['id'], s['network_id'])

    def test_get_sync_data_metering_with_inactive_router(self):
        with self.subnet() as subnet:
            s = subnet['subnet']
            self._set_net_external(s['network_id'])
            with self.router(name='router1') as router1:
                self._add_external_gateway_to_router(
                    router1['router']['id'], s['network_id'])
                with self.router(
                    name='router2', admin_state_up=False
                ) as router2:
                    self._add_external_gateway_to_router(
                        router2['router']['id'], s['network_id'])
                    with self.metering_label():
                        callbacks = metering_rpc.MeteringRpcCallbacks(
                            self.meter_plugin)
                        data = callbacks.get_sync_data_metering(
                            self.adminContext, host='agent1')
                        self.assertEqual(
                            set(['router1']), set([r['name'] for r in data]))

                    self._remove_external_gateway_from_router(
                        router2['router']['id'], s['network_id'])

                self._remove_external_gateway_from_router(
                    router1['router']['id'], s['network_id'])



# Copyright (c) 2019 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import network_segment_range as range_def
from neutron_lib.api import validators
from neutron_lib import constants as const
from neutron_lib.db import api as db_api
from neutron_lib import exceptions as lib_exc
from neutron_lib.exceptions import network_segment_range as range_exc
from neutron_lib.plugins import directory
from neutron_lib.plugins import utils as plugin_utils
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log

from neutron._i18n import _
from neutron.db import segments_db
from neutron.extensions import network_segment_range as ext_range
from neutron.objects import base as base_obj
from neutron.objects import network_segment_range as obj_network_segment_range

LOG = log.getLogger(__name__)


def is_network_segment_range_enabled():
    network_segment_range_class = ('neutron.services.network_segment_range.'
                                   'plugin.NetworkSegmentRangePlugin')
    return any(p in cfg.CONF.service_plugins
               for p in ['network_segment_range', network_segment_range_class])


def _get_physical_network(network_segment_range):
    if network_segment_range.get('network_type') != const.TYPE_VLAN:
        return None

    physical_network = network_segment_range.get(
        "physical_network", const.ATTR_NOT_SPECIFIED)
    if not validators.is_attr_set(physical_network):
        message = _("Network type %s requires 'physical_network' to be "
                    "specified while creating new range") % const.TYPE_VLAN
        raise lib_exc.BadRequest(resource=range_def.RESOURCE_NAME,
                                 msg=message)
    return physical_network


class NetworkSegmentRangePlugin(ext_range.NetworkSegmentRangePluginBase):
    """Implements Neutron Network Segment Range Service plugin."""

    supported_extension_aliases = [range_def.ALIAS]

    __native_pagination_support = True
    __native_sorting_support = True
    __filter_validation_support = True

    def __init__(self):
        super(NetworkSegmentRangePlugin, self).__init__()
        self.type_manager = directory.get_plugin().type_manager
        self.type_manager.initialize_network_segment_range_support()

    def _get_network_segment_range(self, context, id):
        obj = obj_network_segment_range.NetworkSegmentRange.get_object(
            context, id=id)
        if obj is None:
            raise range_exc.NetworkSegmentRangeNotFound(range_id=id)
        return obj

    def _validate_network_segment_range_eligible(self, network_segment_range):
        range_data = (network_segment_range.get('minimum'),
                      network_segment_range.get('maximum'))
        # Currently, network segment range only supports VLAN, VxLAN,
        # GRE and Geneve.
        if network_segment_range.get('network_type') == const.TYPE_VLAN:
            plugin_utils.verify_vlan_range(range_data)
        else:
            plugin_utils.verify_tunnel_range(
                range_data, network_segment_range.get('network_type'))

    def _validate_network_segment_range_overlap(self, context,
                                                network_segment_range):
        filters = {
            'default': False,
            'network_type': network_segment_range['network_type'],
            'physical_network': _get_physical_network(network_segment_range)
        }
        range_objs = obj_network_segment_range.NetworkSegmentRange.get_objects(
            context, **filters)
        overlapped_range_id = [
            obj.id for obj in range_objs if
            (network_segment_range['minimum'] <= obj.maximum and
             network_segment_range['maximum'] >= obj.minimum)]
        if overlapped_range_id:
            raise range_exc.NetworkSegmentRangeOverlaps(
                range_id=', '.join(overlapped_range_id))

    def _add_unchanged_range_attributes(self, updates, existing):
        """Adds data for unspecified fields on incoming update requests."""
        for key, value in existing.items():
            updates.setdefault(key, value)
        return updates

    def _is_network_segment_range_referenced(self, context,
                                             network_segment_range):
        return segments_db.network_segments_exist_in_range(
            context, network_segment_range['network_type'],
            network_segment_range.get('physical_network'),
            network_segment_range)

    def _is_network_segment_range_type_supported(self, network_type):
        if not (self.type_manager.network_type_supported(network_type) and
                network_type in const.NETWORK_SEGMENT_RANGE_TYPES):
            raise range_exc.NetworkSegmentRangeNetTypeNotSupported(
                type=network_type)

        return True

    def _are_allocated_segments_in_range_impacted(self, context,
                                                  existing_range,
                                                  updated_range):
        updated_range_min = updated_range.get('minimum',
                                              existing_range['minimum'])
        updated_range_max = updated_range.get('maximum',
                                              existing_range['maximum'])
        existing_range_min, existing_range_max = (
            segments_db.min_max_actual_segments_in_range(
                context, existing_range['network_type'],
                existing_range.get('physical_network'), existing_range))

        if existing_range_min and existing_range_max:
            return bool(updated_range_min >= existing_range_min or
                        updated_range_max <= existing_range_max)
        return False

    @log_helpers.log_method_call
    def create_network_segment_range(self, context, network_segment_range):
        """Check network types supported on network segment range creation."""
        range_data = network_segment_range['network_segment_range']
        if self._is_network_segment_range_type_supported(
                range_data['network_type']):
            with db_api.CONTEXT_WRITER.using(context):
                self._validate_network_segment_range_eligible(range_data)
                self._validate_network_segment_range_overlap(context,
                                                             range_data)
                network_segment_range = (
                    obj_network_segment_range.NetworkSegmentRange(
                        context, name=range_data['name'],
                        description=range_data.get('description'),
                        default=False,
                        shared=range_data['shared'],
                        project_id=(range_data['project_id']
                                    if not range_data['shared'] else None),
                        network_type=range_data['network_type'],
                        physical_network=(range_data['physical_network']
                                          if range_data['network_type'] ==
                                          const.TYPE_VLAN else None),
                        minimum=range_data['minimum'],
                        maximum=range_data['maximum'])
                )
                network_segment_range.create()

        self.type_manager.update_network_segment_range_allocations(
            network_segment_range['network_type'])
        return network_segment_range.to_dict()

    @log_helpers.log_method_call
    def get_network_segment_range(self, context, id, fields=None):
        network_segment_range = self._get_network_segment_range(
            context, id)
        return network_segment_range.to_dict(fields=fields)

    @log_helpers.log_method_call
    def get_network_segment_ranges(self, context, filters=None, fields=None,
                                   sorts=None, limit=None, marker=None,
                                   page_reverse=False):
        # TODO(kailun): Based on the current spec:
        #  https://review.opendev.org/599980, this method call may
        #  possibly return a large amount of data since ``available``
        #  segment list and ``used`` segment/project mapping will be also
        #  returned and they can be large sometimes. Considering that this
        #  API is admin-only and list operations won't be called often based
        #  on the use cases, we'll keep this open for now and evaluate the
        #  potential impacts. An alternative is to return the ``available``
        #  and ``used`` segment number or percentage.
        pager = base_obj.Pager(sorts, limit, page_reverse, marker)
        filters = filters or {}
        network_segment_ranges = (
            obj_network_segment_range.NetworkSegmentRange.get_objects(
                context, _pager=pager, **filters))

        return [
            network_segment_range.to_dict(fields=fields)
            for network_segment_range in network_segment_ranges
        ]

    @log_helpers.log_method_call
    def update_network_segment_range(self, context, id, network_segment_range):
        """Check existing network segment range impact on range updates."""
        updated_range_data = network_segment_range['network_segment_range']
        with db_api.CONTEXT_WRITER.using(context):
            network_segment_range = self._get_network_segment_range(context,
                                                                    id)
            existing_range_data = network_segment_range.to_dict()

            if existing_range_data['default']:
                raise range_exc.NetworkSegmentRangeDefaultReadOnly(range_id=id)

            if self._are_allocated_segments_in_range_impacted(
                    context,
                    existing_range=existing_range_data,
                    updated_range=updated_range_data):
                raise range_exc.NetworkSegmentRangeReferencedByProject(
                    range_id=id)

            new_range_data = self._add_unchanged_range_attributes(
                updated_range_data, existing_range_data)
            self._validate_network_segment_range_eligible(new_range_data)
            network_segment_range.update_fields(new_range_data)
            network_segment_range.update()

        self.type_manager.update_network_segment_range_allocations(
            network_segment_range['network_type'])
        return network_segment_range.to_dict()

    @log_helpers.log_method_call
    def delete_network_segment_range(self, context, id):
        """Check segment reference on network segment range deletion."""
        with db_api.CONTEXT_WRITER.using(context):
            network_segment_range = self._get_network_segment_range(context,
                                                                    id)
            range_data = network_segment_range.to_dict()

            if range_data['default']:
                raise range_exc.NetworkSegmentRangeDefaultReadOnly(range_id=id)

            if self._is_network_segment_range_referenced(
                    context, range_data):
                raise range_exc.NetworkSegmentRangeReferencedByProject(
                    range_id=id)
            network_segment_range.delete()

        self.type_manager.update_network_segment_range_allocations(
            network_segment_range['network_type'])



# Copyright 2023 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op


"""Create L3HARouterNetwork.project_id unique constraint

Revision ID: 682c319773d7
Revises: 6f1145bff34c
Create Date: 2023-04-27 13:45:05.103963

"""

# revision identifiers, used by Alembic.
revision = '682c319773d7'
down_revision = '6f1145bff34c'


TABLE = 'ha_router_networks'
COLUMN = 'project_id'


def upgrade():
    op.create_unique_constraint(
        constraint_name='uniq_%s0%s' % (TABLE, COLUMN),
        table_name=TABLE,
        columns=[COLUMN])



# Copyright (c) 2023 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import port_hardware_offload_type as phot
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_const

from neutron.objects.port.extensions import port_hardware_offload_type as \
    phot_obj


class PortHardwareOffloadTypeDbMixin(object):
    """Mixin class to add hardware offload type extension to a port"""

    def _process_create_port(self, context, data, result):
        hw_type = data.get(phot.HARDWARE_OFFLOAD_TYPE)
        if hw_type not in n_const.VALID_HWOL_TYPES:
            result[phot.HARDWARE_OFFLOAD_TYPE] = None
            return

        obj = phot_obj.PortHardwareOffloadType(
            context, port_id=result['id'], hardware_offload_type=hw_type)
        obj.create()
        result[phot.HARDWARE_OFFLOAD_TYPE] = hw_type
        # NOTE(ralonsoh): this is updating not the "result" dictionary but
        # the "data" dictionary that are the API input parameters.
        try:
            pb_profile = data[portbindings.PROFILE]
            capabilities = pb_profile.get('capabilities', [])
            if hw_type not in capabilities:
                capabilities.append(hw_type)
            data[portbindings.PROFILE]['capabilities'] = capabilities
        except KeyError:
            data[portbindings.PROFILE] = {'capabilities': [hw_type]}

    def _extend_port_dict(self, port_db, result):
        if port_db.hardware_offload_type:
            result[phot.HARDWARE_OFFLOAD_TYPE] = (
                port_db.hardware_offload_type.hardware_offload_type)
        else:
            result[phot.HARDWARE_OFFLOAD_TYPE] = None



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from neutron_lib.agent import topics
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib import constants
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from neutron_lib.utils import net
from oslo_log import log as logging
import oslo_messaging
from oslo_utils import versionutils

from neutron.api.rpc.callbacks import resources
from neutron.api.rpc.handlers import resources_rpc
from neutron.common import utils
from neutron.db import securitygroups_rpc_base as sg_rpc_base

LOG = logging.getLogger(__name__)


class SecurityGroupServerRpcApi(object):
    """RPC client for security group methods in the plugin.

    This class implements the client side of an rpc interface.  This interface
    is used by agents to call security group related methods implemented on the
    plugin side.  The other side of this interface is defined in
    SecurityGroupServerRpcCallback.  For more information about changing rpc
    interfaces, see doc/source/contributor/internals/rpc_api.rst.
    """
    def __init__(self, topic):
        target = oslo_messaging.Target(
            topic=topic, version='1.0',
            namespace=constants.RPC_NAMESPACE_SECGROUP)
        self.client = n_rpc.get_client(target)

    def security_group_rules_for_devices(self, context, devices):
        LOG.debug("Get security group rules "
                  "for devices via rpc %r", devices)
        cctxt = self.client.prepare(version='1.1')
        return cctxt.call(context, 'security_group_rules_for_devices',
                          devices=devices)

    def security_group_info_for_devices(self, context, devices):
        LOG.debug("Get security group information for devices via rpc %r",
                  devices)
        call_version = '1.3'
        cctxt = self.client.prepare(version=call_version)
        return cctxt.call(context, 'security_group_info_for_devices',
                          devices=devices,
                          call_version=call_version)


class SecurityGroupServerRpcCallback(object):
    """Callback for SecurityGroup agent RPC in plugin implementations.

    This class implements the server side of an rpc interface.  The client side
    can be found in SecurityGroupServerRpcApi. For more information on changing
    rpc interfaces, see doc/source/contributor/internals/rpc_api.rst.
    """

    # API version history:
    #   1.1 - Initial version
    #   1.2 - security_group_info_for_devices introduced as an optimization
    #   1.3 - security_group_info_for_devices returns member_ips with new
    #         structure.

    # NOTE: target must not be overridden in subclasses
    # to keep RPC API version consistent across plugins.
    target = oslo_messaging.Target(version='1.3',
                                   namespace=constants.RPC_NAMESPACE_SECGROUP)

    @property
    def plugin(self):
        return directory.get_plugin()

    def _get_devices_info(self, context, devices):
        return dict(
            (port['id'], port)
            for port in self.plugin.get_ports_from_devices(context, devices)
            if port and not net.is_port_trusted(port)
        )

    def security_group_rules_for_devices(self, context, **kwargs):
        """Callback method to return security group rules for each port.

        also convert remote_group_id rule
        to source_ip_prefix and dest_ip_prefix rule

        :params devices: list of devices
        :returns: port correspond to the devices with security group rules
        """
        devices_info = kwargs.get('devices')
        ports = self._get_devices_info(context, devices_info)
        return self.plugin.security_group_rules_for_ports(context, ports)

    def security_group_info_for_devices(self, context, **kwargs):
        """Return security group information for requested devices.

        :params devices: list of devices
        :returns:
        sg_info{
          'security_groups': {sg_id: [rule1, rule2]}
          'sg_member_ips': {sg_id: {'IPv4': set(), 'IPv6': set()}}
          'devices': {device_id: {device_info}}
        }

        Note that sets are serialized into lists by rpc code.
        """
        # The original client RPC version was 1.2 before this change.
        call_version = kwargs.pop("call_version", '1.2')
        _target_version = versionutils.convert_version_to_tuple(call_version)
        devices_info = kwargs.get('devices')
        ports = self._get_devices_info(context, devices_info)
        sg_info = self.plugin.security_group_info_for_ports(context, ports)
        if _target_version < (1, 3):
            LOG.warning("RPC security_group_info_for_devices call has "
                        "inconsistent version between server and agents. "
                        "The server supports RPC version is 1.3 while "
                        "the agent is %s.", call_version)
            return self.make_compatible_sg_member_ips(sg_info)
        return sg_info

    def make_compatible_sg_member_ips(self, sg_info):
        sg_member_ips = sg_info.get('sg_member_ips', {})
        sg_ids = sg_member_ips.keys()
        for sg_id in sg_ids:
            member_ips = sg_member_ips.get(sg_id, {})
            ipv4_ips = member_ips.get("IPv4", set())
            comp_ipv4_ips = set([ip for ip, _mac in ipv4_ips])
            ipv6_ips = member_ips.get("IPv6", set())
            comp_ipv6_ips = set([ip for ip, _mac in ipv6_ips])
            comp_ips = {"IPv4": comp_ipv4_ips,
                        "IPv6": comp_ipv6_ips}
            sg_member_ips[sg_id] = comp_ips
        sg_info['sg_member_ips'] = sg_member_ips
        return sg_info


class SecurityGroupAgentRpcApiMixin(object):
    """RPC client for security group methods to the agent.

    This class implements the client side of an rpc interface.  This interface
    is used by plugins to call security group methods implemented on the
    agent side.  The other side of this interface can be found in
    SecurityGroupAgentRpcCallbackMixin.  For more information about changing
    rpc interfaces, see doc/source/contributor/internals/rpc_api.rst.
    """

    # history
    #   1.1 Support Security Group RPC
    SG_RPC_VERSION = "1.1"

    def _get_security_group_topic(self):
        return topics.get_topic_name(self.topic,
                                     topics.SECURITY_GROUP,
                                     topics.UPDATE)

    @utils.disable_notifications
    def security_groups_rule_updated(self, context, security_groups):
        """Notify rule updated security groups."""
        if not security_groups:
            return
        cctxt = self.client.prepare(version=self.SG_RPC_VERSION,
                                    topic=self._get_security_group_topic(),
                                    fanout=True)
        cctxt.cast(context, 'security_groups_rule_updated',
                   security_groups=security_groups)

    @utils.disable_notifications
    def security_groups_member_updated(self, context, security_groups):
        """Notify member updated security groups."""
        if not security_groups:
            return
        cctxt = self.client.prepare(version=self.SG_RPC_VERSION,
                                    topic=self._get_security_group_topic(),
                                    fanout=True)
        cctxt.cast(context, 'security_groups_member_updated',
                   security_groups=security_groups)


class SecurityGroupAgentRpcCallbackMixin(object):
    """A mix-in that enable SecurityGroup support in agent implementations.

    This class implements the server side of an rpc interface.  The client side
    can be found in SecurityGroupAgentRpcApiMixin. For more information on
    changing rpc interfaces, see doc/source/contributor/internals/rpc_api.rst.

    The sg_agent reference implementation is available in neutron/agent
    """
    # mix-in object should be have sg_agent
    sg_agent = None

    def _security_groups_agent_not_set(self):
        LOG.warning("Security group agent binding currently not set. "
                    "This should be set by the end of the init "
                    "process.")

    def security_groups_rule_updated(self, context, **kwargs):
        """Callback for security group rule update.

        :param security_groups: list of updated security_groups
        """
        security_groups = kwargs.get('security_groups', [])
        LOG.debug("Security group rule updated on remote: %s",
                  security_groups)
        if not self.sg_agent:
            return self._security_groups_agent_not_set()
        self.sg_agent.security_groups_rule_updated(security_groups)

    def security_groups_member_updated(self, context, **kwargs):
        """Callback for security group member update.

        :param security_groups: list of updated security_groups
        """
        security_groups = kwargs.get('security_groups', [])
        LOG.debug("Security group member updated on remote: %s",
                  security_groups)
        if not self.sg_agent:
            return self._security_groups_agent_not_set()
        self.sg_agent.security_groups_member_updated(security_groups)


class SecurityGroupServerAPIShim(sg_rpc_base.SecurityGroupInfoAPIMixin):
    """Agent-side replacement for SecurityGroupServerRpcApi using local data.

    This provides the same methods as SecurityGroupServerRpcApi but it reads
    from the updates delivered to the push notifications cache rather than
    calling the server.
    """
    def __init__(self, rcache):
        self.rcache = rcache
        registry.subscribe(self._clear_child_sg_rules, 'SecurityGroup',
                           events.AFTER_DELETE)
        registry.subscribe(self._add_child_sg_rules, 'SecurityGroup',
                           events.AFTER_UPDATE)
        # set this attr so agent can adjust the timeout of the client
        self.client = resources_rpc.ResourcesPullRpcApi().client

    def register_legacy_sg_notification_callbacks(self, sg_agent):
        self._sg_agent = sg_agent
        registry.subscribe(self._handle_sg_rule_delete,
                           'SecurityGroupRule', events.AFTER_DELETE)
        registry.subscribe(self._handle_sg_rule_update,
                           'SecurityGroupRule', events.AFTER_UPDATE)
        registry.subscribe(self._handle_sg_member_delete,
                           'Port', events.AFTER_DELETE)
        registry.subscribe(self._handle_sg_member_update,
                           'Port', events.AFTER_UPDATE)
        self._register_legacy_ag_notification_callbacks(sg_agent)

    def _register_legacy_ag_notification_callbacks(self, sg_agent):
        for event in (events.AFTER_UPDATE, events.AFTER_DELETE):
            registry.subscribe(self._handle_address_group_event,
                               resources.ADDRESSGROUP, event)

    def security_group_info_for_devices(self, context, devices):
        ports = self._get_devices_info(context, devices)
        result = self.security_group_info_for_ports(context, ports)
        return result

    def security_group_rules_for_devices(self, context, devices):
        # this is the legacy method that should never be called since
        # security_group_info_for_devices will never throw an unsupported
        # error.
        raise NotImplementedError()

    def get_secgroup_ids_for_address_group(self, address_group_id):
        filters = {'remote_address_group_id': (address_group_id, )}
        return set([rule.security_group_id for rule in
                    self.rcache.get_resources('SecurityGroupRule', filters)])

    def _add_child_sg_rules(self, rtype, event, trigger, payload):
        # whenever we receive a full security group, add all child rules
        # because the server won't emit events for the individual rules on
        # creation.
        context = payload.context
        updated = payload.latest_state
        for rule in updated.rules:
            self.rcache.record_resource_update(context, 'SecurityGroupRule',
                                               rule)

    def _clear_child_sg_rules(self, rtype, event, trigger, payload):
        context = payload.context
        existing = payload.states[0]
        if not existing:
            return
        # the server can delete an entire security group without notifying
        # about the security group rules. so we need to emulate a rule deletion
        # when a security group is removed.

        rules = self.rcache.match_resources_with_func(
            'SecurityGroupRule',
            lambda rule: rule.security_group_id == existing.id)

        for rule in rules:
            self.rcache.record_resource_delete(context, 'SecurityGroupRule',
                                               rule.id)
        # If there's a rule which remote is the deleted sg, remove that also.
        rules = self.rcache.match_resources_with_func(
            'SecurityGroupRule',
            lambda sg_rule: sg_rule.remote_group_id == existing.id)
        for rule in rules:
            self.rcache.record_resource_delete(context, 'SecurityGroupRule',
                                               rule.id)

    def _handle_sg_rule_delete(self, rtype, event, trigger, payload):
        existing = payload.states[0]
        if not existing:
            return
        sg_id = existing.security_group_id
        self._sg_agent.security_groups_rule_updated([sg_id])

    def _handle_sg_rule_update(self, rtype, event, trigger, payload):
        updated = payload.latest_state
        sg_id = updated.security_group_id
        self._sg_agent.security_groups_rule_updated([sg_id])

    def _handle_sg_member_delete(self, rtype, event, trigger, payload):
        # received on port delete
        existing = payload.states[0]
        sgs = set(existing.security_group_ids) if existing else set()
        if sgs:
            self._sg_agent.security_groups_member_updated(sgs)

    def _handle_sg_member_update(self, rtype, event, trigger, payload):
        # received on port update
        existing = payload.states[0]
        updated = payload.latest_state
        changed_fields = payload.metadata['changed_fields']
        sgs = set(existing.security_group_ids) if existing else set()
        if not changed_fields.intersection({'security_group_ids', 'fixed_ips',
                                            'allowed_address_pairs'}):
            # none of the relevant fields to SG calculations changed
            return
        sgs.update(set(updated.security_group_ids))
        if sgs:
            self._sg_agent.security_groups_member_updated(sgs)

    def _handle_address_group_event(self, rtype, event, trigger, payload):
        resource_id = payload.resource_id
        if event == events.AFTER_UPDATE:
            self._sg_agent.address_group_updated(resource_id)
        else:
            self._sg_agent.address_group_deleted(resource_id)

    def _get_devices_info(self, context, devices):
        # NOTE(kevinbenton): this format is required by the sg code, it is
        # defined in get_port_from_device and mimics
        # make_port_dict_with_security_groups in ML2 db
        result = {}
        for device in devices:
            ovo = self.rcache.get_resource_by_id('Port', device)
            if not ovo:
                continue
            port = ovo.to_dict()
            # the caller expects trusted ports to be excluded from the result
            if net.is_port_trusted(port):
                continue

            port['security_groups'] = list(ovo.security_group_ids)
            port['security_group_rules'] = []
            port['security_group_source_groups'] = []
            port['security_group_remote_address_groups'] = []
            port['fixed_ips'] = [str(f['ip_address'])
                                 for f in port['fixed_ips']]
            # NOTE(kevinbenton): this id==device is only safe for OVS. a lookup
            # will be required for linux bridge and others that don't have the
            # full port UUID
            port['device'] = port['id']
            port['port_security_enabled'] = getattr(
                ovo.security, 'port_security_enabled', True)
            result[device] = port
        return result

    def _select_ips_for_remote_group(self, context, remote_group_ids):
        if not remote_group_ids:
            return {}
        ips_by_group = {rg: set() for rg in remote_group_ids}

        filters = {'security_group_ids': tuple(remote_group_ids)}
        for p in self.rcache.get_resources('Port', filters):
            allowed_ips = [(str(addr.ip_address), str(addr.mac_address))
                           for addr in p.allowed_address_pairs]
            port_ips = [(str(addr.ip_address), str(p.mac_address))
                        for addr in p.fixed_ips] + allowed_ips
            for sg_id in p.security_group_ids:
                if sg_id in ips_by_group:
                    ips_by_group[sg_id].update(set(port_ips))
        return ips_by_group

    def _select_ips_for_remote_address_group(self, context,
                                             remote_address_group_ids):
        if not remote_address_group_ids:
            return {}
        ips_by_group = {ag: set() for ag in remote_address_group_ids}
        for ag_id in remote_address_group_ids:
            ag = self.rcache.get_resource_by_id('AddressGroup', ag_id)
            # NOTE(hangyang) Expected data structure is (ip, mac) tuple
            ips = [(str(addr_assoc.address), None)
                   for addr_assoc in ag.addresses]
            ips_by_group[ag_id].update(set(ips))
        return ips_by_group

    def _select_rules_for_ports(self, context, ports):
        if not ports:
            return []
        results = []
        sg_ids = set((sg_id for p in ports.values()
                      for sg_id in p['security_group_ids']))
        rules_by_sgid = collections.defaultdict(list)
        for sg_id in sg_ids:
            filters = {'security_group_id': (sg_id, )}
            for r in self.rcache.get_resources('SecurityGroupRule', filters):
                rules_by_sgid[r.security_group_id].append(r)
        for p in ports.values():
            for sg_id in p['security_group_ids']:
                for rule in rules_by_sgid[sg_id]:
                    results.append((p['id'], rule.to_dict()))
        return results

    def _select_sg_ids_for_ports(self, context, ports):
        sg_ids = set((sg_id for p in ports.values()
                      for sg_id in p['security_group_ids']))
        return [(sg_id, ) for sg_id in sg_ids]

    def _get_sgs_stateful_flag(self, context, sg_ids):
        sgs_stateful = {}
        for sg_id in sg_ids:
            sg = self.rcache.get_resource_by_id(resources.SECURITYGROUP, sg_id)
            sgs_stateful[sg_id] = sg.stateful

        return sgs_stateful



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import availability_zone as az_def
from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.api import extensions
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib.db import resource_extend
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory

from neutron.common import utils
from neutron.db import l3_attrs_db


@resource_extend.has_resource_extenders
@registry.has_registry_receivers
class RouterAvailabilityZoneMixin(l3_attrs_db.ExtraAttributesMixin):
    """Mixin class to enable router's availability zone attributes."""

    @staticmethod
    @resource_extend.extends([l3_apidef.ROUTERS])
    def _add_az_to_response(router_res, router_db):
        l3_plugin = directory.get_plugin(constants.L3)
        if not extensions.is_extension_supported(
                l3_plugin, 'router_availability_zone'):
            return
        router_res['availability_zones'] = (
            l3_plugin.get_router_availability_zones(router_db))

    @registry.receives(resources.ROUTER, [events.PRECOMMIT_CREATE])
    def _process_az_request(self, resource, event, trigger, payload):
        context = payload.context
        router = payload.latest_state
        router_db = payload.metadata['router_db']
        az_hints = utils.get_az_hints(router)
        if az_hints:
            self.validate_availability_zones(context, 'router',
                                             az_hints)
            self.set_extra_attr_value(router_db, az_def.AZ_HINTS, az_hints)



# Copyright 2021 Troila
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from webob import exc

from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.api.definitions import l3_conntrack_helper as l3_ct
from oslo_utils import uuidutils

from neutron.extensions import l3
from neutron.extensions import l3_conntrack_helper
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.extensions import test_l3

_uuid = uuidutils.generate_uuid


class TestL3ConntrackHelperServicePlugin(test_l3.TestL3NatServicePlugin):
    supported_extension_aliases = [l3_apidef.ALIAS, l3_ct.ALIAS]


class ExtendL3ConntrackHelperExtensionManager(object):

    def get_resources(self):
        return (l3.L3.get_resources() +
                l3_conntrack_helper.L3_conntrack_helper.get_resources())

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class L3NConntrackHelperTestCase(test_l3.L3BaseForIntTests,
                                 test_l3.L3NatTestCaseMixin):
    tenant_id = _uuid()
    fmt = "json"

    def setUp(self):
        mock.patch('neutron.api.rpc.handlers.resources_rpc.'
                   'ResourcesPushRpcApi').start()
        svc_plugins = ('neutron.services.conntrack_helper.plugin.Plugin',
                       'neutron.tests.unit.extensions.'
                       'test_l3_conntrack_helper.'
                       'TestL3ConntrackHelperServicePlugin')
        plugin = ('neutron.tests.unit.extensions.test_l3.TestL3NatIntPlugin')
        ext_mgr = ExtendL3ConntrackHelperExtensionManager()
        super(L3NConntrackHelperTestCase, self).setUp(
              ext_mgr=ext_mgr, service_plugins=svc_plugins, plugin=plugin)
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)

    def _create_router_conntrack_helper(self, fmt, router_id,
                                        protocol, port, helper):
        data = {'conntrack_helper': {
            "protocol": protocol,
            "port": port,
            "helper": helper}
        }
        router_ct_req = self.new_create_request(
            'routers', data,
            fmt or self.fmt, id=router_id,
            subresource='conntrack_helpers',
            as_admin=True)

        return router_ct_req.get_response(self.ext_api)

    def _update_router_conntrack_helper(self, fmt, router_id,
                                        conntrack_helper_id, **kwargs):
        conntrack_helper = {}
        for k, v in kwargs.items():
            conntrack_helper[k] = v
        data = {'conntrack_helper': conntrack_helper}

        router_ct_req = self.new_update_request(
            'routers', data, router_id,
            fmt or self.fmt, sub_id=conntrack_helper_id,
            subresource='conntrack_helpers', as_admin=True)
        return router_ct_req.get_response(self.ext_api)

    def test_create_ct_with_duplicate_entry(self):
        with self.router() as router:
            ct1 = self._create_router_conntrack_helper(
                self.fmt, router['router']['id'],
                "udp", 69, "tftp")
            self.assertEqual(exc.HTTPCreated.code, ct1.status_code)
            ct2 = self._create_router_conntrack_helper(
                self.fmt, router['router']['id'],
                "udp", 69, "tftp")
            self.assertEqual(exc.HTTPBadRequest.code, ct2.status_code)
            expect_msg = ("Bad conntrack_helper request: A duplicate "
                          "conntrack helper entry with same attributes "
                          "already exists, conflicting values are "
                          "{'router_id': '%s', 'protocol': 'udp', "
                          "'port': 69, 'helper': "
                          "'tftp'}.") % router['router']['id']
            self.assertEqual(
                expect_msg, ct2.json_body['NeutronError']['message'])

    def test_update_ct_with_duplicate_entry(self):
        with self.router() as router:
            ct1 = self._create_router_conntrack_helper(
                self.fmt, router['router']['id'],
                "udp", 69, "tftp")
            self.assertEqual(exc.HTTPCreated.code, ct1.status_code)
            ct2 = self._create_router_conntrack_helper(
                self.fmt, router['router']['id'],
                "udp", 68, "tftp")
            self.assertEqual(exc.HTTPCreated.code, ct2.status_code)
            result = self._update_router_conntrack_helper(
                self.fmt, router['router']['id'],
                ct1.json['conntrack_helper']['id'],
                **{'port': 68})
            self.assertEqual(exc.HTTPBadRequest.code, result.status_code)
            expect_msg = ("Bad conntrack_helper request: A duplicate "
                          "conntrack helper entry with same attributes "
                          "already exists, conflicting values are "
                          "{'router_id': '%s', 'protocol': 'udp', "
                          "'port': 68, 'helper': "
                          "'tftp'}.") % router['router']['id']
            self.assertEqual(
                expect_msg, result.json_body['NeutronError']['message'])



# Copyright 2016 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

import netaddr

from neutron.tests.common.exclusive_resources import ip_address
from neutron.tests.common.exclusive_resources import resource_allocator


def _get_random_network(low, high, netmask):
    ip = ip_address.get_random_ip(low, high)
    return str(netaddr.IPNetwork("%s/%s" % (ip, netmask)).cidr)


class ExclusiveIPNetwork(resource_allocator.ExclusiveResource):
    """Allocate a non-overlapping ip network.

    :ivar network: allocated ip network
    :type network: netaddr.IPNetwork
    """

    def __init__(self, low, high, netmask):
        super(ExclusiveIPNetwork, self).__init__(
            'ip_networks',
            functools.partial(_get_random_network, low, high, netmask),
            self.is_valid)

    def _setUp(self):
        super(ExclusiveIPNetwork, self)._setUp()
        self.network = netaddr.IPNetwork(self.resource)

    def is_valid(self, new_resource, allocated_resources):
        new_ipset = netaddr.IPSet([new_resource])
        allocated_ipset = netaddr.IPSet(allocated_resources)
        return new_ipset.isdisjoint(allocated_ipset)
